from .common import check_zero, factorials, nCr, convert_indices
from .poly import jetpoly

def sum_by_jetpoly(ls):
    # sum over the members of a given list so that each member e of jetpoly class appears on the
    # left. This can be used to prevent that in case of numpy the new values are numpy.arrays.
    return sum([e for e in ls if isinstance(e, jetpoly)]) + sum([e for e in ls if not isinstance(e, jetpoly)])

def general_leibniz_rule(f1, f2):
    '''
    Compute the higher-order derivatives of the product of two functions f1*f2.
    In the following denote by f^k the k-th derivative of f, evaluated at a specific point.
    It is assumed that len(f1) = len(f2) =: n.
    
    Parameters
    ----------
    f1: list
        List containing the values f1^k for k = 0, ..., n - 1.
    f2: list
        List containing the values f2^k for k = 0, ..., n - 1.
        
    Returns
    -------
    list
        List containing the values (f1*f2)^k for k = 0, ..., n - 1.
    '''
    nmax = len(f1) - 1 # len(f1): max number of summands
    nok = nCr(nmax)
    return [sum_by_jetpoly([nok[n][k]*f1[n-k]*f2[k] if isinstance(f1[n-k], jetpoly) else nok[n][k]*f2[k]*f1[n-k] for k in range(n + 1)]) for n in range(nmax + 1)]

def faa_di_bruno(f, g):
    '''
    Consider the composition f o g of two functions and for h in {f, g} denote by h^k its k-th derivative. Then
    (f o g)^1 = (f^1 o g) g^1
    (f o g)^2 = (f^2 o g) (g^1)**2 + (f^1 o g) g^2
    (f o g)^3 = ... and so on.
    
    Let F = [f o g, (f^1 o g), (f^2 o g), ...] and G = [g, g^1, g^2, ...] be given. Then this
    function will compute (f o g)^k up to the given (common) order, i.e. the
    chain rule for the k-th derivative.
    
    This routine hereby will make use of the equation
    (f^n o g) = sum_(k = 1)^n f^k B(n, k)(g^1, g^2, ..., g^(n - k + 1)) ,  
    where B(n, k) for n >= k are the incomplete exponential Bell polynomials.
    
    Parameters
    ----------
    f: list
        List containing the values F (see above).
    g: list
        List containing the values G (see above).
        
    Returns
    -------
    list
        List containing the values (f o g)^k for k = 0, ..., len(f) - 1.
    '''
    bell = bell_polynomials(*g[1:])
    return [sum_by_jetpoly([bell.get((n, k), 0)*f[k] for k in range(len(f))]) for n in range(len(f))] # or with numpy arrays: np.matmul(bell, f)

def bell_polynomials(*z):
    '''
    Compute the incomplete exponential Bell polynomials B(n, k, [z1, ..., z_{n - k + 1}]) for all 1 <= k <= n.

    Parameters
    ----------
    z: Values at which the Bell polynomials are supposed to be evaluated. 
       Note that n will be assumed to be the number of different z values.

    Returns
    -------
    dict
        dict of Bell polynomials up to the given order n. Its keys denote the indices (n, k) of the polynomials.
    '''
    n = len(z)
    B = {}
    B[(0, 0)] = 1.0
    nok = nCr(n)
    for jn in range(n + 1):
        for jk in range(1, jn + 1):
            B[(jn, jk)] = sum_by_jetpoly([nok[jn - 1][m - 1]*B.get((jn - m, jk - 1), 0)*z[m - 1] for m in range(1, jn - jk + 2)])
    return B
    
class jet:
    def __init__(self, *values, **kwargs):
        self.__array_priority__ = 1000 # prevent numpy __mul__ and force self.__rmul__ instead if multiplication from left with a numpy object
        self.set_array(*values, **kwargs)
        self.graph = kwargs.get('graph', [(1, '0'), self]) # for keeping track of the computation graph
        
    def get_array(self, **kwargs):
        n = kwargs.get('n', self.order)
        return [self.array(k) for k in range(n + 1)]
    
    def set_array(self, *values, **kwargs):
        omax = len(values) - 1
        if omax > 0:
            zero = values[0]*0
        else:
            zero = 0
        self.array = lambda k: values[k] if k <= omax else zero
        self.set_order(n=kwargs.get('n', omax))
            
    def set_order(self, n):
        self.order = n
        self._factorials = factorials(self.order)
        
    def __len__(self):
        return self.order + 1
    
    def __iter__(self): 
        # called whenever a jet is used in a for loop. If this routine does not exists, self.__getitem__ is used (which may not stop)
        return iter(self.get_array())
    
    def __getitem__(self, n):
        # called whenver a jet is subscribed in the form jet[index].
        result = self.__class__(n=self.order, graph=[(1, f'[{n}]'), self.graph])
        result.array = lambda k: self.array(k)[n]
        return result
        
    def __setitem__(self, index, value):
        current_array = self.get_array()
        current_array[index] = value
        self.set_array(*current_array)
                
    def __str__(self):
        outstr = ''
        for e in self.get_array():
            if not isinstance(e, type(self)):
                outstr += e.__str__()
            else:
                outstr += str(e)
            outstr += ', '
        return f'{self.order}-jet({outstr[:-2]})'

    def _repr_html_(self):
        return f'<samp>{self.__str__()}</samp>'
    
    def __neg__(self):
        result = self.__class__(n=self.order, graph=[(1, '-'), self.graph])
        result.array = lambda n: -self.array(n)
        return result
    
    def __add__(self, other):
        other = self.convert(other)
        max_order = max([self.order, other.order])
        result = self.__class__(n=max_order, graph=[(2, '+'), self.graph, other.graph])
        result.array = lambda n: self.array(n) + other.array(n)
        return result
    
    def __radd__(self, other):
        other = self.convert(other)
        return self + other
    
    def __sub__(self, other):
        return self + -other
    
    def __rsub__(self, other):
        return -self + other
        
    def __mul__(self, other):
        other = self.convert(other)
        max_order = max([self.order, other.order])
        result = self.__class__(n=max_order, graph=[(2, '*'), self.graph, other.graph])
        # compute the derivatives      
        f1, f2 = self.get_array(n=max_order), other.get_array(n=max_order)
        glr = general_leibniz_rule(f1, f2)
        result.array = lambda n: glr[n] if n <= max_order else 0
        # result.array = lambda k: general_leibniz_rule(f1[:k + 1], f2[:k + 1]) if k <= max_order else 0
        # The next line would work for arbitrary orders, but it is also much slower instead of pre-loading the array:
        # result.array = lambda n: sum([n_over_k(n, k)*self.array(n - k)*other.array(k) for k in range(n + 1)])
        return result
    
    def __rmul__(self, other):
        other = self.convert(other)
        return self*other
    
    def __pow__(self, other):
        
        other = self.convert(other)
        assert other.order == 0 # only scalars are accepted here.
        m = other.array(0)
        f = self.array(0)
        n = self.order
                
        # compute the maximal number of derivatives of f**m
        for nmax in range(n + 1):
            if m - nmax == 0:
                break
        # N.B. nmax <= self.order
        
        # compute the coefficients in the series [f**m, m*f**(m - 1), m*(m - 1)*f**(m - 2), ...]
        coeff = 1
        h = m
        factors = []
        for j in range(nmax + 1):
            factors.append(coeff)
            coeff *= h
            h -= 1
            
        # compose the above series with the current jet
        result = self.__class__(*[f**(m - k)*factors[k] for k in range(nmax + 1)], n=self.order, graph=[(2, '**'), self.graph, other.graph])
        return result@self
    
    def __rpow__(self, other):
        other = self.convert(other)    
        return other**self
    
    def __truediv__(self, other):
        other = self.convert(other)
        return self.__mul__(other.inv())
    
    def __rtruediv__(self, other):
        other = self.convert(other)
        return other.__mul__(self.inv())
    
    def copy(self):
        '''
        Return a (shallow) copy of this jet.
        '''
        result = self.__class__(n=self.order, graph=self.graph)
        result.array = self.array
        return result
    
    def truncate(self, max_power):
        result = self.__class__(n=self.order, graph=self.graph)
        def result_array(n):
            try:
                return self.array(n).truncate(max_power)
            except:
                return self.array(n)
        result.array = result_array
        return result
    
    def derive(self, **kwargs):
        '''
        Return the derivative of this jet, determined by shifting its array to the left by 1.
        
        Parameters
        ----------
        index: int, optional
            if an index is provided, then the individual entries of the jet are derived with respect
            to this index, if they are of class jetpoly.
        '''
        array = self.get_array()
        result = self.copy()
        if len(array) == 1:
            result.set_array(0)
        else:
            result.set_array(*array[1:])
        return result
    
    def __matmul__(self, other):
        '''
        Compute the composition of two jets, based on Faà di Bruno's formula.
        It is assumed that the array ar with ar = self.get_array() belongs to the values
        [f o g, f^1 o g, f^2 o g, ... ], where f^k denote the k-th derivative of the function f,
        so this function will compute the values
        [f o g, (f o g)^1, (f o g)^2, ...]
        '''        
        f = self.get_array()
        g = other.get_array()
        return self.__class__(*faa_di_bruno(f, g), n=self.order, graph=[(2, 'o'), self.graph, other.graph])
    
    def inv(self):
        '''
        Compute 1/jet. This will be done by applying the chain rule to
        "1/x o jet".
        '''
        f = self.array(0)
        assert not check_zero(f), 'Division by zero at requested point.'

        facts = self._factorials
        invf = [(-1)**n*facts[n]/f**(n + 1) for n in range(self.order + 1)]
        result = self.copy()
        result.set_array(*invf)
        result = result@self
        result.graph = [(1, '1/'), self.graph]
        return result
            
    def __eq__(self, other):
        if isinstance(self, type(other)):
            if self.order != other.order:
                return False
            else:
                return all([check_zero(self.array(k) - other[k]) for k in range(self.order + 1)])
        elif self.order > 1:
            return all([check_zero(self.array(k) - other) for k in range(self.order + 1)])
        else:
            return check_zero(self.array(0) - other)
            
    def convert(self, other, n=0):
        '''
        Convert an object to an instance of this class.
        '''
        if not isinstance(self, type(other)):
            return self.__class__(other, n=n)
        else:
            return other
        
    def conjugate(self):
        # N.B.: (f^(n)).conjugate() = (f.conjugate())^(n)
        result = self.__class__(n=self.order, graph=[(1, 'cg'), self.graph])
        result.array = lambda n: self.array(n).conjugate()
        return result
    
    def real(self):
        return (self + self.conjugate())/2
    
    def imag(self):
        return (self - self.conjugate())/2/1j
    
    def __call__(self, *z):
        result = []
        for k in range(len(self)):
            ek = self.array(k)
            if hasattr(ek, '__call__'):
                result.append(ek(*z))
            else:
                result.append(ek)
        return self.__class__(*result, n=self.order, graph=[(1, '(*)'), self.graph])
    
    def taylor_coefficients(self, n_args: int=0, **kwargs):
        '''Extract the Taylor coefficients from a given jet-evaluation (the output of njet.ad.derive.eval).
        
        Let m be the number of arguments of the involved poylnomials. 
        Then the k-th derivative of f has the form
        (D^k f)(z1, ..., zm) = sum_{j1 + ... + jm = k} Df[j1, ... jm]/(j1! * ... * jm!) * z1**j1 * ... * zm**jm
        = sum_{S(j1, ... jm), j1 + ... + jm = k} C(j1, ..., jm) * Df[j1, ..., jm] * z1**j1 * ... * zm**jm.
        with S(j1, ..., jm) an index denoting the set having the elements {j1, ..., jm},
        Df[j1, ..., jm] := \\partial^j1/\\partial_{z1}^j1 ... \\partial^jm/\\partial_{zm}^jm f (z1, ..., zm)
        and combinatorial factor
        C(j1, ..., jm) = (j1 + ... + jm)!/(j1! * ... * jm!) .
        
        !!! Attention
        The current jet should have jetpoly entries in the k-th order, k > 0.
        
        Parameters
        ----------
        n_args: int
            The total number of involved variables.
            
        **kwargs
            Optional arguments passed to poly.taylor_coefficients routine.
            
            Note that one can control how to deal with multiplicities C(j1, ..., jm) (notation see above) by
            passing mult_drv and mult_prm attributes to this routine.
            
        Returns
        -------
        dict
            Dictionary which maps the tuples representing the indices and powers of the individual
            monomials to their coefficients, corresponding to the Taylor expansion of the given expression.
        '''
        # Use the longer factorials list; this is required if nested derivatives of different orders are used.
        facts = kwargs.get('facts', self._factorials)
        if len(facts) < len(self._factorials):
            facts = self._factorials
        kwargs['facts'] = facts
        
        # Get the actual Taylor-coefficients
        tc = {}
        # Add constants to tc; this needs to be done separately, because often it may happen that the first
        # entry of a jet is just a float, so that these constants will be missed in the next loop.
        self_array = self.get_array()
        const = self_array[0]
        if not check_zero(const):
            if not hasattr(const, 'taylor_coefficients'):
                tc[(0,)*n_args] = const
            else:
                tc.update(const.taylor_coefficients(n_args=n_args, **kwargs))
        
        for entry in self_array[1:]: # iteration over the derivatives of order >= 1.
            if not hasattr(entry, 'taylor_coefficients'): # skip any non-polynomial entry.
                continue
            tc.update(entry.taylor_coefficients(n_args=n_args, **kwargs))
            # Since we loop over derivatives of a specific order, it is ensured that these Taylor coefficients are always different, 
            # so the above procedure does not overwrite existing keys.
        self.tc = tc
        return tc
    
    def extract(self, k: int, **kwargs):
        '''
        Extract the components of the k-th derivative. 
        
        Parameters
        ----------
        k: int
            The order of the derivatives to extract.
            
        tc: dict, optional
            The output of self.taylor_coefficients, containing the entries of the derivative. If nothing
            specified, the last evaluation (stored in self.tc) will be used.
            
        Returns
        -------
        dict
            The components of the k-th derivative of self.func.
        '''
        if 'tc' in kwargs.keys():
            self.tc = kwargs['tc']
        if not hasattr(self, 'tc'):
            self.tc = self.taylor_coefficients(**kwargs)
        return {j: self.tc[j] for j in self.tc.keys() if sum(j) == k}
    
    def build_tensor(self, k: int, **kwargs):
        '''
        Convert the components of the k-th derivative into the entries of a (self.n_args)**k tensor.
        See also self.convert_indices for details.
        
        Parameters
        ----------
        k: int
            The degree of the derivatives to be considered.
        
        Returns
        -------
        dict
            Dictionary representing non-zero components of the tensor which describes the multivariate 
            Taylor map of order k.
            Only values unequal to zero will be stored, and only one member for each set of entries
            which can be obtained by suitable permutations of the indices.
        '''
        entries = self.extract(k=k, **kwargs)
        list_of_tuples = list(entries.keys())
        converted_indices = convert_indices(list_of_tuples)
        return {converted_indices[k]: entries[list_of_tuples[k]] for k in range(len(list_of_tuples))}
    
    def grad(self, **kwargs):
        '''
        Returns the gradient of the function.
        
        Parameters
        ----------
        kwargs: dict
            Additional arguments passed to self.build_tensor
            
        Returns
        -------
        dict
            Dictionary containing the components of the gradient.
        '''
        return self.build_tensor(k=1, **kwargs)
    
    def hess(self, **kwargs):
        '''
        Returns the Hessian of the function.
        
        Parameters
        ----------
        kwargs: dict
            Additional arguments passed to self.build_tensor
            
        Returns
        -------
        dict
            Dictionary containing the components of the Hessian.
        '''
        return self.build_tensor(k=2, **kwargs)


'''    
    njet: A leightweight automatic differentiation library for 
          higher-order automatic differentiation
    
    Copyright (C) 2021, 2022, 2023 by Malte Titze

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''

from ._version import __version__
from .ad import jet, derive, taylor_coefficients
from .jet import jetpoly



# Collection of elementary functions whose derivatives are known to any order.

import numpy
import sympy
import mpmath

def get_function(code: str, name: str):
    '''
    Return the function to handle objects of a specific code.
    
    Parameters
    ----------
    code: str
        The name of the code. Currently supported codes are: 'numpy', 'mpmath',
        'sympy', 'njet'.
        
    name: str
        The name of the function to be returned. See the source code of this
        module for a list of supported functions. 
    '''
    
    sympy_dict = {'sin': sympy.sin,
                  'cos': sympy.cos,
                  'sinh': sympy.sinh,
                  'cosh': sympy.cosh,
                  'exp': sympy.exp,
                  'log': sympy.log}
    
    numpy_dict = {'sin': numpy.sin,
                  'cos': numpy.cos,
                  'sinh': numpy.sinh,
                  'cosh': numpy.cosh,
                  'exp': numpy.exp,
                  'log': numpy.log}
    
    mpmath_dict = {'sin': mpmath.sin,
                  'cos': mpmath.cos,
                  'sinh': mpmath.sinh,
                  'cosh': mpmath.cosh,
                  'exp': mpmath.exp,
                  'log': mpmath.log}
    
    jet_dict = {'sin': sin,
                'cos': cos,
                'sinh': sinh,
                'cosh': cosh,
                'exp': exp,
                'log': log}
    
    if code == 'numpy':
        return numpy_dict[name]
    elif code == 'mpmath':
        return mpmath_dict[name]
    elif code == 'sympy':
        return sympy_dict[name]
    elif code == 'njet':
        return jet_dict[name]
    else:
        return numpy_dict[name] # fall-back to numpy as default
    
def get_package_name(*x):
    '''
    Routine intended to get the package name of a specific object.
    
    Parameters
    ----------
    x: obj
        The object to be examined.
        
    Returns
    -------
    str
        A string denoting the code to be used on the object. 
    '''
    return str(x[0].__class__.__mro__[0].__module__).split('.')[0]
    
def jetfunc(func):
    '''
    General wrapper containing the treatment of input
    which are no njets.
    '''
    def inner(x, **kwargs):
        code = kwargs.get('code', get_package_name(x))
        name = func.__name__
        if code != 'njet':
            return get_function(code, name)(x)
        else:
            # The next lines are always executed,
            # because the built-in functions need to be applied to the
            # 0-th component of the jet:
            x0 = x.array(0)
            code_x0 = get_package_name(x0)
            f = get_function(code_x0, name)
            fx0 = f(x0)
            return func(x, x0=x0, code_x0=code_x0, f=f, fx0=fx0)
    return inner

def zero(x, **kwargs):
    code = kwargs.get('code', get_package_name(x))
    if code == 'numpy':
        return numpy.zeros(x.shape)
    elif code == 'mpmath':
        return x*0
    elif code == 'sympy':
        return x*0
    elif code == 'njet':
        return x*0
    else:
        return x*0

@jetfunc    
def sin(x, **kwargs):
    '''
    Compute the sin of a jet.

    Parameters
    ----------
    x: jet

    Returns
    -------
    jet
    '''
    s = kwargs.get('fx0')
    c = get_function(kwargs.get('code_x0'), 'cos')(kwargs.get('x0'))
    result = x.__class__(s, n=x.order, graph=[(1, 'sin'), x.graph])
    # compute the derivatives
    result.array = lambda n: [s, c, -s, -c][n%4]
    return result@x

@jetfunc
def cos(x, **kwargs):
    '''
    Compute the cos of a jet.

    Parameters
    ----------
    x: jet

    Returns
    -------
    jet
    '''
    c = kwargs.get('fx0')
    s = get_function(kwargs.get('code_x0'), 'sin')(kwargs.get('x0'))
    result = x.__class__(c, n=x.order, graph=[(1, 'cos'), x.graph])
    # compute the derivatives
    result.array = lambda n: [c, -s, -c, s][n%4]
    return result@x

@jetfunc
def exp(x, **kwargs):
    '''
    Compute the exponential of a jet.

    Parameters
    ----------
    x: jet

    Returns
    -------
    jet
    '''
    e = kwargs.get('fx0')
    result = x.__class__(e, n=x.order, graph=[(1, 'exp'), x.graph])
    # compute the derivatives
    result.array = lambda n: e
    return result@x

@jetfunc
def log(x, **kwargs):
    '''
    Compute the natural logarithm of a jet.

    Parameters
    ----------
    x: jet

    Returns
    -------
    jet
    '''
    ln = kwargs.get('fx0')
    # compute the derivatives
    dx = x.copy().derive()
    drx_arr = (dx/x).get_array()[:-1]
    result = x.__class__(n=x.order, graph=[(1, 'log'), x.graph])
    result.set_array(ln, *drx_arr)
    return result

@jetfunc
def sinh(x, **kwargs):
    '''
    Compute the sinh of a jet.

    Parameters
    ----------
    x: jet

    Returns
    -------
    jet
    '''
    sh = kwargs.get('fx0')
    ch = get_function(kwargs.get('code_x0'), 'cosh')(kwargs.get('x0'))
    result = x.__class__(sh, n=x.order, graph=[(1, 'sinh'), x.graph])
    # compute the derivatives
    result.array = lambda n: [sh, ch][n%2]
    return result@x

@jetfunc
def cosh(x, **kwargs):
    '''
    Compute the cosh of a jet.

    Parameters
    ----------
    x: jet

    Returns
    -------
    jet
    '''
    ch = kwargs.get('fx0')
    sh = get_function(kwargs.get('code_x0'), 'sinh')(kwargs.get('x0'))
    result = x.__class__(ch, n=x.order, graph=[(1, 'cosh'), x.graph])
    # compute the derivatives
    result.array = lambda n: [ch, sh][n%2]
    return result@x




import warnings
from .common import check_zero, factorials

class jetpoly:
    '''
    Implementation of a polynomial with arbitrary number of variables.
    
    The information of the coefficients is stored internally as dictionary in 
    the 'terms' field. One can use the 'terms' field as keyworded argument in order
    to construct a custom polynomial.

    A polynomial p can be initiated with a value v and (optional) its index (describing
    the variable) and power n so that overall p = v*x_i**n. A general polynomial in several
    variables can then be constructed by successive addition and multiplication of different 
    classes with each other.
    
    Example
    -------    
    | self.terms = {
    | frozenset({(0, 1)}): 11,
    | frozenset({(0, 4), (1, 22), (13, 1)}): 89,
    | frozenset({(4, 9), (7, 2), (12, 33)}): -0.22
    | }
    
    | The interpretation is:
    | 11*x0**1 +
    | 89*x0**4*x1**22*x13**1 +
    | -0.22*x4**9*x7**2*x12**33
    '''
    def __init__(self, value=0, index: int=0, power: int=0, **kwargs):
        self.__array_priority__ = 1000 # prevent numpy __mul__ and force self.__rmul__ instead if multiplication from left with a numpy object
        if 'terms' in kwargs.keys():
            self.terms = kwargs['terms']
        elif not check_zero(value):
            self.terms = {frozenset([(index, power)]): value}
        else:
            self.terms = {}
            
    def truncate(self, max_power):
        '''
        Drop all terms beyond a specific power.
        
        Parameters
        ----------
        max_power: int
            The maximal power beyond which we want to drop terms.
            
        Returns
        -------
        jetpoly
            A jetpoly object of power up (and including) max_power.
        '''
        new_terms = {}
        for k, v in self.terms.items():
            max_power_k = sum([e[1] for e in k])
            if max_power_k <= max_power:
                new_terms[k] = v            
        return self.__class__(terms=new_terms)
        
    def __str__(self):
        if len(self.terms) == 0:
            return '0'
        else:
            outstr = '['
            for key, value in self.terms.items():
                fac = ''
                for e in key:
                    if e[1] == 0: # do not show z**0
                        continue
                    fac += f'*x{e[0]}**{e[1]}'
                outstr += f'{value}]{fac} + \n ['
            return outstr[:-6]
    
    def _repr_html_(self):
        outstr = self.__str__().replace('\n', '<br>')
        return f'<samp>{outstr}</samp>'
    
    def __add__(self, other):
        if not isinstance(self, type(other)):
            other = self.__class__(other)
        new_terms = {}
        for k in set(self.terms).union(set(other.terms)):
            new_value = self.terms.get(k, 0) + other.terms.get(k, 0)
            if check_zero(new_value): # do not store 0-terms; x may be float or numpy array etc.
                continue
            new_terms[k] = new_value
        if len(new_terms) == 0:
            return self.__class__(0)
        else:
            return self.__class__(terms=new_terms)
    
    def __radd__(self, other):
        other = self.__class__(other)
        return other + self
    
    def __neg__(self):
        new_terms = {}
        for key, value in self.terms.items():
            new_terms[key] = -value
        return self.__class__(terms=new_terms)
    
    def __sub__(self, other):
        return self + -other
    
    def __rsub__(self, other):
        return other + -self
    
    def __mul__(self, other):
        # Interpretation: (sum_j a_j)*(sum_k b_k) = sum_{j, k} a_j*b_k
        if not isinstance(self, type(other)):
            return self.__class__(terms={k: v*other for k, v in self.terms.items()})
        pol_prod = {}
        for aj, value1 in self.terms.items():
            e1 = dict(aj) # e.g. e1 = {0: 0, 1: 5, 2: 3}
            se1 = set(e1)
            for bk, value2 in other.terms.items():
                value_prod = value1*value2
                e2 = dict(bk) # e.g. e2 = {0: 0, 1: 0, 2: 1}
                e_prod = []
                for k in se1.union(set(e2)):
                    power = e1.get(k, 0) + e2.get(k, 0)
                    if power > 0:
                        e_prod.append((k, power))
                if len(e_prod) == 0:
                    e_prod = frozenset({(0, 0)})
                else:
                    e_prod = frozenset(e_prod)
                pol_prod[e_prod] = value_prod + pol_prod.get(e_prod, 0) # account for multiplicity
        pol_prod = {k: v for k, v in pol_prod.items() if not check_zero(v)} # remove zero values
        if len(pol_prod) == 0: # in this case zero(s) are produced
            return self.__class__(0)
        else:
            return self.__class__(terms=pol_prod)

    def __rmul__(self, other):
        other = self.__class__(other)
        return other*self
    
    def __truediv__(self, other):
        assert not check_zero(other)
        return self.__class__(terms={k: v/other for k, v in self.terms.items()})
    
    def __pow__(self, other):
        assert type(other) == int
        assert other >= 0
        if other == 0:
            return self.__class__(1) # N.B. 0**0 := 1

        remainder = other%2
        half = self**(other//2)
        if remainder == 1:
            return self*half*half
        else:
            return half*half
        
    def __eq__(self, other):
        if not isinstance(self, type(other)):
            return self.terms == self.__class__(other).terms
        else:
            return self.terms == other.terms
        
    def __getitem__(self, n):
        return self.__class__(terms={k: v[n] for k, v in self.terms.items() if not check_zero(v[n])})
        
    def conjugate(self):
        r'''
        Conjugate the terms of the current polynomial.
        
        Consider the following example:
        Let f = u + 1j*v be a complex function in which u and v are real functions from
        R^2 -> R^2. Denote by Df := \partial f/\partial z and D'f := \partial f/\partial \bar z
        the Wirtinger operators and for any g = x + 1j*y the complex conjugation g' = x - 1j*y.
        Then we have for the differential of f:
        
        df = Df*dz + D'f*dz',
        
        (df)' = D(f')*dz + D'(f')*dz'.
        
        Since our polynomials represent differentials, in this example df plays 
        the role of a polynomial of the two independent variables dz and dz' (the variables z and z' behave
        independent when applying the Wirtinger operators to any function g(z, z'), as one can show). 
        This means that in the complex case conjugation should be propagated to the underlying 
        function: f --> f', without changing the differentials themselves (the keys of self.terms).
        
        However, this requires that the keys (related to the variables) are prepared in advance accordingly:
        Every variable needs his complex-conjugate partner, and in the original expression complex conjugation
        needs to be replaced by this partner variable.
        
        Returns
        -------
        jetpoly
        '''
        # Consider the following example:
        new_terms = {}
        for key, value in self.terms.items():
            new_terms[key] = value.conjugate()
        return self.__class__(terms=new_terms)
    
    def real(self):
        return (self + self.conjugate())/2
    
    def imag(self):
        return (self - self.conjugate())/2/1j
        
    def taylor_coefficients(self, n_args: int=0, facts=[], mult_prm: bool=True, mult_drv: bool=True):
        '''
        Obtain the Taylor coefficients of the current polynomial.
        
        Parameters
        ----------
        n_args: int, optional
            The total number of involved parameters. If nothing specified, the number of involved variables
            is determined from the current indices of the polynomial.
            
        facts: list, optional
            A list containing the factorial numbers up to the maximal order in the current polynomial.
            Hereby it must hold facts[k] = k!
            
        mult_prm: bool, optional
            Whether or not to include multiplicities into the final result related to the permulation of expressions (e.g. derivatives)
            (default: True).
            
        mult_drv: bool, optional
            Whether or not to include multiplicities related to the derivatives of powers (default: True)
        
        Returns
        -------
        dict
            A dictionary of the form "tuple: value", where "tuple" consists of series of powers, hereby the entry
            j denotes the power of the j-th variable.
        '''
        if n_args == 0 or len(facts) == 0:
            # determine the number of arguments from the maximal index of the polynomial.
            order, indices_set = self.get_order(return_indices=True)
            try:
                n_args = max([n_args, max(indices_set) + 1])
            except:
                # in this case indices_set was empty, so the polynomial power is 0
                warnings.warn("'n_args' parameter recommended.")
                return {}
            if len(facts) == 0:
                facts = factorials(order)
                
        taylor_coeffs = {}
        for key, value in self.terms.items(): # loop over the individual polynomials of the k-th derivative
            # key corresponds to a specific frozenset, i.e. some indices and powers of a specific monomial.
            
            # Step 1: Construct the indices containing the powers of the variables at their respective positions.
            # If there is no index with a power > 0, then 'indices' will not be changed. Also determine
            # the multiplicity.
            multiplicity = 1
            indices = [0]*n_args
            for index, power in key:
                # index: the index of the variable; power: its respective power
                if power == 0: # the (k, 0)-entries correspond to the scalar 1 and will be ignored here.
                    continue
                indices[index] = power
                if mult_prm or mult_drv:
                    multiplicity = multiplicity*facts[power]
                    
            # Step 2: Update the Taylor-coefficients dictionary
            if not check_zero(value): # only add non-zero values
                if mult_drv: # remove the factorials in the Taylor expansion, here related to the derivatives of the powers.
                    value = value*multiplicity
                if mult_prm:
                    value = value/facts[sum(indices)] # the denominator ensures to remove multiplicities emerging from permutations of derivatives.

                tpl = tuple(indices)
                taylor_coeffs[tpl] = taylor_coeffs.get(tpl, 0) + value
                
        return taylor_coeffs
        
    def __call__(self, *z):
        '''
        Evaluate the current polynomial at a specific point.
        
        Note that a sufficient number of parameters must be given, according to the
        indices of the variables the polynomial represents. For example, a polynomial dependent
        on "x4" has to be provided with at least 4 parameters.
        '''
        result = 0
        for fs, v in self.terms.items():
            f = 1
            for index, power in fs:
                f *= z[index]**power
            result += f*v
        return result
          
    def get_order(self, return_indices=False):
        '''
        Return the order of the current polynomial.
        
        Parameters
        ----------
        return_indices: boolean, optional
            If true, also return the indices of the current polynomial.
            
        Returns
        -------
        int
            The order.
            
        set
            A set of indices as described above.
        '''
        ps = 0
        args = set()
        for monomial_basis in self.terms.keys():
            monomial_power = 0
            for index, power in monomial_basis:
                monomial_power += power
                args.add(index)
            ps = max([ps, monomial_power])
        if return_indices:
            return ps, args
        else:
            return ps
        
        


import numpy as np
from more_itertools import distinct_permutations, windowed
from tqdm import tqdm
from copy import copy
import warnings

from . import jet, jetpoly, derive, taylor_coefficients
from .common import factorials, check_zero

def accel_asc(n: int):
    '''
    Compute the partitions of a given integer.
    
    Parameters
    ----------
    n: int
    
    Returns
    -------
    generator
        A generator, producing lists of integers which sum up to n. Every
        list hereby represents a unique set of integers.
    '''
    # taken from https://stackoverflow.com/questions/10035752/elegant-python-code-for-integer-partitioning
    a = [0 for i in range(n + 1)]
    k = 1
    y = n - 1
    while k != 0:
        x = a[k - 1] + 1
        k -= 1
        while 2 * x <= y:
            a[k] = x
            y -= x
            k += 1
        l = k + 1
        while x <= y:
            a[k] = x
            a[l] = y
            yield a[:k + 2]
            x += 1
            y -= 1
        a[k] = x + y
        y = x + y - 1
        yield a[:k + 1]

def symtensor_call(z, terms: dict):
    '''
    Let 'terms' represent the terms of a *homogeneous* jetpoly object 
    P = sum_{j1, ..., jr} P_{j1, ..., jr} * x_1^{n_1} * ... * x_r^{n_r}
    of degree k (i.e. k = j1 + ... + jr).
    
    Then this routine will compute, for a given set of k vectors z = [z1, ..., zk]
    the result
    P[z] = sum_{j1, ..., jr} P_{j1, ..., jr} * z1_1^{n_1} * ... * zk_r^{n_r},
    where zj_w denotes the w-th component of the vector zj.
    
    Parameters
    ----------
    z: list
        A list of iterables, where each element of z represents a vector of the same dimension.
        
    terms: dict
        A dictionary representing the terms of a homogeneous jetpoly object. 
        Attention: No check will be made if the jetpoly object is actually homogeneous.
        
    Returns
    -------
    jetpoly
        A jetpoly object, which is the result of inserting the given vectors into 
        the polynomial, as described above.
    '''
    # Further information can be found here:
    # https://mathoverflow.net/questions/106323/faa-di-brunos-formula-for-vector-valued-functions
    dim = len(z)
    result = 0
    for fs, v in terms.items():
        f = 1
        k = 0
        for index, power in fs:
            for p in range(power):
                f *= z[k][index]
                k += 1
        result += f*v
    return result

def _make_run_params(max_order: int):
    '''
    Compute the summation indices for a given order and load them into memory.
    This routine is intended to be used in order to avoid re-calculating 
    these indices every time the general Faa di Bruno routine is called (for a specific
    max_order).
    
    Parameters
    ----------
    max_order: int
        The requested maximal order up to which the derivatives should be computed.
        
    Returns
    -------
    list
        A list of tuples, where each tuple T = (order, r, [n1, ..., nr]) represents the
        order number, the length r of the partition [n1, ..., nr] of order and the partition itself.
    '''
    facts = factorials(max_order)
    summation_indices = [(order, len(j), dp) for order in range(1, max_order + 1) for j in accel_asc(order) for dp in distinct_permutations(j)]
    return {'factorials': facts, 'indices': summation_indices}

def general_faa_di_bruno(f, g, run_params={}):
    '''
    Faa di Bruno for vector-valued functions.
    
    Let G: K^l -> K^m and F: K^m -> K^w be two vector-valued functions. The
    goal is to compute the higher-order derivatives of the composition F o G.
    
    Assume that f = [f1, ..., fw] and g = [g1, ..., gm] represent the 
    n-jet-collections of F and G, i.e. fk = n-jet(fk_1, ..., fk_n)
    where fk_j represents the j-th derivative of the k-th component of F etc.
    
    Note that this routine requires that both f and g are represented by jetpoly
    objects in their higher-order array entries (they store the various partial derivatives
    for the given orders).
    In particular, the ordinary derive.eval routine will produce such objects. 
    This is the major difference to the one-dimensional Faa di Bruno formula 
    in njet.jet.faa_di_bruno, which has no such requirement.
    
    Parameters
    ----------
    f: list
        A list of n-jet objects, representing the derivatives of the function F
        at position G(z).
        
    g: list
        A list of n-jet objects, representing the derivatives of the function G at
        a given position z.
        
    run_params: dict, optional
        A dictionary containing the output of njet.extras._make_run_params. These parameters
        can be send to the routine to avoid internal re-calculation when the routine
        it is called repeatedly.
    
    Returns
    -------
    list
        A list of jet objects, representing the higher-orders of the components
        of the compositional map F o G.
    '''
    # Further information can be found here:
    # https://mathoverflow.net/questions/106323/faa-di-brunos-formula-for-vector-valued-functions
    n_dim = len(g) # (== domain dimension of f)
    max_order = f[0].order # number of requested derivatives
    assert all([jf.order == max_order for jf in f] + [jg.order == max_order for jg in g])
    
    if len(run_params) == 0:
        run_params = _make_run_params(max_order)
    facts, indices = run_params['factorials'], run_params['indices']
    
    out = [[fk.array(0) for fk in f]] + [[0 for k in range(n_dim)] for l in range(max_order)]
    for order, r, e in indices:
        for k in range(n_dim):
            jfk = f[k]
            if check_zero(jfk.array(r)): 
                # skip in case that the r-th derivative of the k-th component of jfk does not exist
                continue
            out[order][k] += symtensor_call([[jg.array(nj)/facts[nj] for jg in g] for nj in e], jfk.array(r).terms)/facts[r]*facts[order]
    return [jet(*[out[k][j] for k in range(max_order + 1)], n=max_order) for j in range(n_dim)]

def compose(*evals, run_params={}, **kwargs):
    '''
    Compose derivatives of a chain of vector-valued jet-evaluations.

    Parameters
    ----------
    evals: 
        A series of (multi-dimensional) jet-evaluations. Hereby every jet-evaluation 
        is represented by a list of n-jets, one for each individual component.

    **kwargs: dict, optional
        One can pass the boolean parameter disable_tqdm to enable (if False) a progress bar.

    Returns
    -------
    list
        A list of jet evaluations for each vector component of the chain, representing
        the Taylor-expansion of the chain.
    '''
    if len(run_params) == 0:
        max_order = max([e.order for f in evals for e in f])
        run_params = _make_run_params(max_order)
    
    evr = evals[0]
    out = [evr]
    for k in tqdm(range(1, len(evals)), disable=kwargs.get('disable_tqdm', True)):
        evr = general_faa_di_bruno(evals[k], evr, run_params=run_params)
        out.append(evr)
    return out

#################################
#   Jet manipulation routines   #
#################################

def tile(jev, ncopies: int):
    '''
    Construct a new jet in which its entries contain copies of itself.
    
    Intended to work if the jet entries are numpy arrays. A jet with
    scalar values can be converted to a jet containing numpy arrays
    of shape (1,) by using ncopies=1.
    
    Parameters
    ----------
    jev: jet
    
    ncopies: int
        Number of desired copies.
    
    Returns
    -------
    jet
        A jet having arrays of length ncopies + 1
    '''
    assert ncopies >= 0
    if ncopies == 0:
        return jev
    new_jet_array = []
    for k in range(jev.order + 1):
        jk = jev.array(k)
        if hasattr(jk, 'terms'): # Assume the entry is of type jetpoly
            entry_k = jev.array(k).terms
            new_terms_k = {}
            for key, value in entry_k.items():
                new_terms_k[key] = np.tile(value, [ncopies] + [1]*len(value.shape))
            new_jet_array.append(jetpoly(terms=new_terms_k))
        else:
            new_jet_array.append(np.tile(jk, [ncopies] + [1]*len(jk.shape)))
    return jet(*new_jet_array)

def _jetp1(jev, index: int, to_concat0=None, ncopies: int=1, location=-1):
    '''
    Add copies of a unity transformations to the give jet entries -- either
    in the end or at the beginning.
    
    Intended to work if all jet entries are numpy arrays.

    Parameters
    ----------
    jev: jet
    
    index: int
        The component of the unity transformation.
        
    to_concat0: array-like, optional
        The value to be concatenated, representing the 0-entry of the jet array.
        If nothing specified, the last (or first) value of jetv.array(0) will be used.
        
    ncopies: int, optional
        The number of copies attached.
        
    location: int, optional
        -1: Attach to end. Everything else: attach to start.
    
    Returns
    -------
    jet
    '''
    assert ncopies >= 0
    if ncopies == 0:
        return jev
    
    a0 = jev.array(0)
    zero = a0*0
    one = zero + 1
    kj = frozenset({(index, 1)})
    if to_concat0 is None:
        if location == -1:
            to_concat0 = [a0[-1]]*ncopies # The index '-1' here is related to the use of _jetp1 in the cderive.cycle routine: In case there are several passages through the specific function, we will take a copy of the last passage. This also means that _jetp1 is intended to be used on jets whose entries carry numpy arrays.
        else:
            to_contat0 = [a0[0]]*ncopies
        
    # the unity transformation will reproduce the value of the jet in the 0-array.
    if location == -1:
        #a0_new = np.r_[a0, a0[-1]]
        a0_new = np.concatenate([a0, to_concat0])
    else:
        a0_new = np.concatenate([to_concat0, a0])
    
    new_jet_array = [a0_new]
    for k in range(1, jev.order + 1):
        jk = jev.array(k)
        if not hasattr(jk, 'terms'):
            entry_k = {}
        else:
            entry_k = jev.array(k).terms
        if k == 1:
            # ensure that kj appears in the new jetpoly in any case:
            _ = entry_k.setdefault(kj, zero)
        new_terms_k = {}
        for key, value in entry_k.items():
            if key == kj: # may happen if k = 1
                to_concatj = [one[0]]*ncopies
            else:
                to_concatj = [zero[0]]*ncopies
                
            if location == -1:
                #new_terms_k[key] = np.r_[value, 1]
                new_terms_k[key] = np.concatenate([value, to_concatj])
            else:
                #new_terms_k[key] = np.r_[value, 0]
                new_terms_k[key] = np.concatenate([to_concatj, value])
        new_jet_array.append(jetpoly(terms=new_terms_k))
    return jet(*new_jet_array)

def _jbuild(jets):
    '''
    Combine the arrays of the given jets, to return a new jet.
    Internal routine; all jetpoly objects need to have identical keys.
    
    Parameters
    ----------
    jets: list
        A list of njet.jet objects.
        
    Returns
    -------
    jet
        A jet in which the entries of each original jet are included in a large numpy array.
    '''
    max_order = max([j.order for j in jets])
    new_array = [np.array([j.array(0) for j in jets])]
    for k in range(1, max_order + 1):
        new_terms = {}
        all_keys = set().union(*[j.array(k).terms.keys() for j in jets])
        for key in all_keys:
            new_terms[key] = np.array([j.array(k).terms.get(key, 0) for j in jets])
        new_array.append(jetpoly(terms=new_terms))
    return jet(*new_array, n=max_order)

##########################
#   Chain-derive class   #
##########################

class cderive:
    '''
    Class to handle the derivatives of a chain of vector-valued functions with repetitions. 
    The given functions should be unique, while their repetition in the chain
    will be given by an optional ordering.

    Parameters
    ----------
    functions: list
        A list of callable(s) or 'derive' classes related to the
        unique vector-valued*) functions in the chain to be derived.

        Alternatively, one can pass a list of 'derive' objects, containing
        vector-valued jet evaluation output. 
        This second case can be used to avoid re-calculating the derivatives
        in circumstances where it is not required.

        Attention:
        By default, the first function is assumed to be executed first, in contrast
        to the mathematical notation. This can be changed by passing
        an 'ordering' argument (see below).

        *) This means that the functions must return iterables in any case.

    order: int, optional
        The maximal order of the derivatives to be computed. If nothing specified,
        the first derivative(s) will be computed.

    ordering: list, optional
        The order defining how the unique functions are arranged in the chain.
        Hereby the index j must refer to the function at position j in the sequence
        of functions.

    run_params: dict, optional
        A dictionary containing the output of njet.extras._make_run_params. These parameters
        can be send to the routine to avoid internal re-calculation when the routine
        it is called repeatedly.
        
    reset: boolean, optional
        Parameter given to self.set_ordering. If 'True' (default), erase any jet-evaluation data 
        coming along with the input (the ._evaluation fields). To prevent this behavior, 
        set this parameter to 'False'. However, in this case one should ensure that correct 
        data has been stored -- in particular concerning the requested ordering of the chain.

    **kwargs
        Optional keyworded arguments passed to njet.ad.derive init.
    '''

    def __init__(self, *functions, order: int=1, ordering=None, run_params={}, reset=True, **kwargs):

        # Determine user input for the 'functions' parameter
        supported_objects = ['njet.ad.derive', 'njet.extras.cderive'] # objects of these kinds will not be instantiated with 'derive'
        self.dfunctions = [copy(f) if any([s in f.__repr__() for s in supported_objects]) else derive(copy(f), order=order, **kwargs) for f in functions]
        
        self.n_functions = len(self.dfunctions)
        self.order = order
        
        if len(run_params) == 0:
            run_params = _make_run_params(order)
        self.run_params = run_params
        
        self.set_ordering(ordering=ordering, reset=reset)
                
    def set_ordering(self, ordering=None, reset=True):
        '''
        Set the ordering of the current chain. Also compute the number of passages expected through
        each element according to the ordering.
        
        Parameters
        ----------
        ordering: list, optional
            The order defining how the unique functions are arranged in the chain.
            Hereby the index j must refer to the function at position j in the sequence
            of functions.
            
        reset: boolean, optional
            Reset the jet-evaluations as well. This may become necessary because these evaluations
            depend strongly on the current ordering.
        '''
        if ordering is None:
            ordering = list(range(self.n_functions))
        assert len(ordering) > 0, 'Ordering list empty.'
            
        # Check input consistency
        uord = np.unique(ordering).tolist()
        assert len(uord) == self.n_functions and uord[0] == 0, 'Number of functions not consistent with the unique items in the ordering.'

        self.ordering = [k for k in ordering] # copy to prevent modification of input somewhere else
        # For every element in the chain, note a number at which a point
        # passing through the given chain will pass through the element.
        # So for example, self.path[k] = [j1, j2, j3, ...] means that
        # the first passage through element k occurs at global position j1,
        # the second passage through element k occurs at global position j2 etc.
        self.path = {k: [] for k in uord}
        for j in range(len(self)):
            self.path[self.ordering[j]].append(j)

        if not reset:
            # check at least if the (new) path is consistent with the given data:
            _ = self._check_path_consistency()
        else:
            self.reset()
        
    def _check_path_consistency(self):
        '''
        Check if the number of entries for each evaluation (if they exist) is consistent with the internal path enumeration.
        '''
        result = True
        for k in range(len(self)):
            f = self.dfunctions[self.ordering[k]]
            if not hasattr(f, '_evaluation'):
                continue
            index = self.path[self.ordering[k]].index(k)
            if index == 0: # Single passages or possible scalar values will be ignored
                continue
            if not all([index < len(e.array(0)) for e in f._evaluation]): # More data than required will also be ignored.
                warnings.warn(f'Insufficient jet-evaluation entries at position >= {k}. Reset or re-evaluation required.')
                result = False
                break
        return result
                
    def reset(self):
        '''
        Remove all jet evaluations (if they exist).
        '''
        for f in self.dfunctions:
            if hasattr(f, '_evaluation'):
                delattr(f, '_evaluation')
                            
    def jetfunc(self, *point, **kwargs):
        '''
        Run a point through the chain.
        
        Parameters
        ----------
        *point: single value or array-like
            The point at which the chain should be evaluated.
            
        **kwargs: dict, optional
            Keyworded arguments passed to the underlying jet-functions.
            
        Returns
        -------
        single value or array-like
            The final value after the chain of functions has been traversed.
        '''
        self._input = point
        out = [point]
        for k in range(len(self)):
            point = self.dfunctions[self.ordering[k]].jetfunc(*point, **kwargs)
            out.append(point)
        self._output = out
        return point
    
    def _probe(self, *point, **kwargs):
        '''
        Check if the points in the current output are in agreement with the one stored in the input functions.
        This check requires that the input functions have been evaluated and a probe run has been performed.
        
        Parameters
        ----------
        *point: single value or array-like
            The point at which the chain should be probed.
            
        **kwargs: dict, optional
            Keyworded arguments passed to the underlying jet-functions.
            
        Returns
        -------
        boolean
            If True, the currently stored data along the chain will be produced
            by collecting the transverse points starting with the given point through the chain.
        '''
        if hasattr(self, '_input'):
            check = all([check_zero(point[k] - self._input[k]) for k in range(len(point))])
            if check == False:
                return False
            
        if not all([hasattr(df, '_input') for df in self.dfunctions]):
            # Nothing can be compared, so the check will fail
            return False
            
        if not hasattr(self, '_output'): 
            # Probe the current chain
            _ = self.jetfunc(*point, **kwargs)
            
        # Check the points along the chain
        try:
            return all([np.array([self.dfunctions[self.ordering[k]]._input[component_index][self.path[self.ordering[k]].index(k)].array(0) == self._output[k][component_index] for component_index in range(len(self._output[k]))]).all() for k in range(len(self))])
        except:
            warnings.warn('Probe check aborted with error(s).')
            return False
                    
    def eval(self, *point, compose=True, **kwargs):
        '''
        Evaluate the functions in the chain at the requested point.
        
        Parameters
        ----------
        *point: single value or array-like
            The point at which the evaluation should be performed.
            
        compose: boolean, optional
            If true (default), compose the evaluations after they have been computed and 
            return the result, resembling the behaviour of the original njet.derive.eval routine.
            If false, only generate the jet-evaluation data (e.g. in preparation for a succeeding merge command).
            
        **kwargs: dict, optional
            Keyworded arguments passed to the underlying jet-functions.
            
            One can pass the boolean parameter disable_tqdm to enable (if False) a progress bar.
            
        Returns
        -------
        list
            The outcome of self.compose routine (containing the jet-evaluations for each component).
        '''        
        _ = self.jetfunc(*point, **kwargs)
        points_at_functions = [[self._output[l] for l in range(len(self)) if self.ordering[l] == k] for k in range(self.n_functions)]
        # let Q = points_per_function[j], so Q is a list of points which needs to be computed for function j
        # Then the first element in Q is the one which needs to be applied first, etc. (for element j) by this construction.
        
        for k in tqdm(range(self.n_functions), disable=kwargs.get('disable_tqdm', True)):
            function = self.dfunctions[k].jetfunc
            n_args_function = self.dfunctions[k].n_args
            points_at_function = points_at_functions[k]
            components = [np.array([points_at_function[j][l] for j in range(len(points_at_function))], dtype=np.complex128) for l in range(n_args_function)]
            _ = self.dfunctions[k].eval(*components, **kwargs)
            
        if compose:
            return self.compose(**kwargs)
    
    def jev(self, pos: int):
        '''
        Convenience function to obtain jet-evaluation data at a specific position
        (such data can be produced by the 'eval' command).
        
        Parameters
        ----------
        pos: integer
            The position within the chain of functions.
            
        Returns
        -------
        list
            A list of jet-evaluations.
        '''
        return [e[self.path[self.ordering[pos]].index(pos)] for e in self[pos]._evaluation]
    
    def compose(self, **kwargs):
        '''
        Compose the current jet-evaluations in the chain (requires function evaluations in advance).
        
        See njet.extras.compose for details; its output is stored in self._evaluation._chain, while the last element is returned.
        '''
        assert all(hasattr(f, '_evaluation') for f in self.dfunctions), 'Composition requires function evaluations in advance.'
        # the path through the chain will require the selection of the point through which the trajectory passes at the respective position:
        evals = [[e[self.path[self.ordering[pos]].index(pos)] for e in self[pos]._evaluation] for pos in range(len(self))]
        self._evaluation_chain = compose(*evals, run_params=self.run_params, **kwargs)
        self._evaluation = self._evaluation_chain[-1]
        return self._evaluation
    
    def _eval_memcheck(self, *z, **kwargs):
        '''
        Check internal memory if evaluation is required.
        
        Parameters
        ----------
        *z
            The point at which the evaluation should be calculated.
            
        **kwargs
            Additional keyworded parameters intended for the functions in the chain.
        
        Returns
        -------
        eval_required: boolean
            True, if a (re-)evaluation is required.
            
        kwargs_changed: boolean
            True, if the keyworded arguments have been changed in comparison to a previous run.
        '''
        # Determine if the keyworded arguments have been changed
        kwargs_changed = True
        if hasattr(self, '_call_kwargs'):
            kwargs_changed = self._call_kwargs != kwargs
        if kwargs_changed or not hasattr(self, '_call_kwargs'):
            self._call_kwargs = kwargs
            
        # Determine if a (re-)evaluation is required
        eval_required = False
        if len(z) > 0:
            if kwargs_changed:
                eval_required = True
            if not all([hasattr(df, '_evaluation') for df in self.dfunctions]):
                eval_required = True
            elif not self._probe(*z, **kwargs):
                eval_required = True
        return eval_required, kwargs_changed

    def __call__(self, *z, **kwargs):
        '''
        Compute the derivatives of the chain of functions at a given point.
        
        Parameters
        ----------
        *z: single value or array-like
            The point at which the derivative(s) should be evaluated.
            
        **kwargs: dict, optional
            Keyworded arguments passed to the underlying jet functions.
            
        Returns
        -------
        dict
            The Taylor-coefficients of the chain at the point of interest.
        '''
        # These two keywords are reserved for the taylor_coefficients routine and will be removed from the input:
        mult_prm = kwargs.pop('mult_prm', True)
        mult_drv = kwargs.pop('mult_drv', True)
               
        # Perform the composition, if necessary
        eval_required, kwargs_changed = self._eval_memcheck(*z, **kwargs)
        if eval_required:
            _ = self.eval(*z, **kwargs) # evaluation includes 'self.compose'
        elif not hasattr(self, '_evaluation') or kwargs_changed:
            _ = self.compose(**kwargs)

        return taylor_coefficients(self._evaluation, n_args=self.dfunctions[0].n_args, mult_prm=mult_prm, mult_drv=mult_drv)

    def merge(self, pattern=(), positions=None, **kwargs):
        '''
        Merge one or more sections in the current chain simultaneously.

        If a pattern may occur on several places, an
        additional parameter 'positions' has to be provided, so that
        all patterns are non-overlapping.
        
        Parameters
        ----------
        pattern: tuple, optional
            Tuple of integers which defines a subsequence in self.ordering.
            If nothings specified, the entire sequence will be used, and so
            this routine becomes very similar to self.compose (with the difference that
            a cderive object will be returned here).
            
        positions: list, optional
            List of integers which defines the start indices of the above pattern in self.ordering.
            If nothing specified, every occurence of 'pattern' in self.ordering will be used.
            
            In any case, only the members of 'pattern' are merged internally, so the number
            of occurences of that pattern should not affect the computational cost.
            
        **kwargs
            Optional keyworded arguments passed to the individual derive class(es).
            
        Returns
        -------
        cderive
            A cderive object which contains a sequence of 'derive' classes, representing a new chain
            of functions in which the selected pattern(s) have been merged.
        '''
        # Input handling and consistency checks
        #######################################
        assert type(pattern) == tuple
        if len(pattern) == 0:
            pattern = tuple(self.ordering)
        size = len(pattern)
        assert 2 <= size and size <= len(self)
        if positions is None:
            positions = _get_pattern_positions(self.ordering, pattern)
            if len(positions) == 0:
                raise RuntimeError('Pattern not found in the sequence.')
        # Sections must not overlap & can be found in the ordering. Evaluation(s) must exist.
        n_patterns = len(positions)
        assert 0 <= min(positions) and max(positions) < len(self) - size + 1, 'Pattern positions out of bounds.'
        for k in range(n_patterns - 1):
            assert positions[k + 1] - positions[k] >= size, 'Overlapping pattern.'
        assert all(tuple(self.ordering[pos: pos + size]) == pattern for pos in positions), 'Not all patterns found in sequence at requested positions.'
        assert all(hasattr(self.dfunctions[k], '_evaluation') for k in pattern), 'Merging requires function evaluations in advance.'
        self._merge_pattern = pattern
        self._merge_positions = positions

        # Merge the members of the pattern
        ##################################
        passage_indices = [[self.ordering[:pos + k].count(pattern[k]) for k in range(size)] for pos in positions] # The number of passages the functions already had in the chain, before the respective pattern(s).
        evr = [e[[passage_indices[k][0] for k in range(n_patterns)]] for e in self.dfunctions[pattern[0]]._evaluation] # The start values of the merged element consists of all start values at the various positions of the pattern
        for k in tqdm(range(1, size), disable=kwargs.get('disable_tqdm', True)):
            ev = [e[[passage_indices[j][k] for j in range(n_patterns)]] for e in self.dfunctions[pattern[k]]._evaluation] # (**)
            evr = general_faa_di_bruno(ev, evr, run_params=self.run_params)

        # Determine the new ordering & path
        ###################################
        # We also need to modify the evaluation results of the other elements in the chain (at (+) below).
        # This is required, because having merged some of its elements, some evaluations are not valid/required anymore.
        # For example, if we merge element Z=XYX in a chain of the form AXBXYX, then the element X originally occured 3 times. So any point will pass 3 times through that
        # chain. However, after merging, that element X will be present only once in the new lattice AXBZ, so only one entry remains.

        # Define a composition function representing the pattern.
        def pattern_function(*z, **pkwargs):
            for k in range(size):
                f = self.dfunctions[pattern[k]].jetfunc
                z = f(*z, **pkwargs)
            return z
        dp = derive(pattern_function, order=self.order, n_args=getattr(self.dfunctions[self.ordering[k]], 'n_args', kwargs.get('n_args', 0)))            
        dp._evaluation = evr # Note that 'evr' contains only those points which are passing through the pattern, which is guaranteed at step (**) above.
                
        pattern_positions = [r for pos in positions for r in range(pos, pos + size)]
        placeholder = -1
        ordering_w_placeholder = [self.ordering[j] if j not in pattern_positions else placeholder for j in range(len(self))]
        k = 0
        new_functions = []
        unique_function_indices = [] # to ensure that we pic only the unique functions in the chain
        while k < len(self):
            no = ordering_w_placeholder[k]
            if no == placeholder:
                if k == positions[0]: # first occurence of the pattern
                    new_functions.append(dp)
                    unique_function_indices.append(placeholder)
                k += size
            else:
                func_index = ordering_w_placeholder[k]
                
                if func_index in unique_function_indices:
                    k += 1
                    continue

                func = self.dfunctions[func_index]
                if func_index in pattern:
                    # (+) The function 'original_func' is in the pattern, but has not been merged (because it is not covered by a pattern within the sequence).
                    # In this case we have to remove those points in the function evaluation(s) which belong to passages through the (now merged) pattern(s).
                    func_path = self.path[func_index]
                    evaluation = [e[[j not in pattern_positions for j in func_path]] for e in func._evaluation]
                    new_func = func.__class__(func.jetfunc, n_args=func.n_args, order=func.order) # initiate (copy) the original function in order to not overwrite its "._evaluation" field.
                    new_func._evaluation = evaluation
                else:
                    new_func = func

                new_functions.append(new_func)
                unique_function_indices.append(func_index)
                k += 1
                
        # remove the chain of placeholders in the new_ordering & recalculate the ordering
        new_ordering = []
        j = 0
        while j < len(self):
            if ordering_w_placeholder[j] == placeholder:
                new_ordering.append(placeholder)
                j += size
            else:
                new_ordering.append(ordering_w_placeholder[j])
                j += 1
        new_ordering = _get_ordering(new_ordering)
        
        return self.__class__(*new_functions, order=self.order, ordering=new_ordering, run_params=self.run_params, reset=False)
    
    def __len__(self):
        return len(self.ordering)
    
    def __getitem__(self, key):
        '''
        If a list is provided, return an object of type(self).
        Otherwise, return the individual derive/cderive object.
        
        Attention: There is only support for strictly monotonous increasing lists
                   without gaps, as the individual jet-evaluations may become invalid
                   otherwise.
        '''
        if type(key) == list:
            requested_func_indices = [self.ordering[e] for e in key]
            requested_positions = key
            if len(key) > 1:
                # If the order has been changed, this may lead to wrong evaluations (they do not 'commute')
                # in this case we shall raise a warning. We shall check this here:
                if not all(i + 1 == j for i, j in zip(key, key[1:])):
                    raise RuntimeError('Requested slice needs to be strictly increasing without any gaps.')        
        else:
            requested_func_indices = self.ordering[key]
            requested_positions = range(len(self))[key]
            
        if type(requested_func_indices) != list:
            return self.dfunctions[requested_func_indices]
        else:
            if len(requested_func_indices) == 0:
                return []
            
            requested_unique_func_indices = list(np.unique(requested_func_indices))
            requested_funcs = [copy(self.dfunctions[e]) for e in requested_unique_func_indices]
            
            # Since we are going to return a cderive object, we remove any evaluation
            # points which belong to elements which are not present in the new chain
            for k in range(len(requested_funcs)):
                rf = requested_funcs[k]
                if not hasattr(rf, '_evaluation'):
                    continue
                findex = requested_unique_func_indices[k]
                func_path = self.path[findex]
                evaluation = [e[[j in requested_positions for j in func_path]] for e in rf._evaluation]
                rf._evaluation = evaluation # n.b. object was copied, so no danger to overwrite one of its fields
            
            new_ordering = [requested_unique_func_indices.index(e) for e in requested_func_indices] # starts from zero up to length of the unique (new) elements
            return self.__class__(*requested_funcs, ordering=new_ordering, order=self.order, run_params=self.run_params, reset=False)
            
    def __iter__(self):
        self._iterpointer = 0
        return self
            
    def __next__(self):
        if self._iterpointer < len(self):
            self._iterpointer += 1
            return self.dfunctions[self.ordering[self._iterpointer - 1]]
        else:
            raise StopIteration
            
    def index(self, value):
        '''
        Return the element number at a given position in the beamline chain (similar as .index for lists)
        '''
        return self.dfunctions.index(value)
    
    def _prepare_cycle(self, *point, periodic=False, warn=False, **kwargs):
        '''
        Internal routine to process user input, check if the current chain admits data 
        suitable for cycling and prepare a function to obtain the jet-evaluations depending on 
        the position in the chain.
                
        Parameters
        ----------
        *point: float or iterable, optional
            The point at which to evaluate the derivative(s). It has to be provided if the chain has
            not yet stored any jet-evaluation data.
        
        periodic: boolean, optional
            If True, assume that the final point equals the given point. In particular, the
            data taken in the calculation will be taken from the current jet-evaluations (if
            it exists).
            
        warn: boolean, optional
            Show warnings.
            
        **kwargs
            Optional keyworded arguments passed to the underlying jet-functions in case an evaluation
            is required. If they differ from self._call_kwargs (if it exists), a warning will be issued.
        
        Returns
        -------
        callable
            A function taking an integer (pos) and returning the jet-evaluation at this position. It will
            have a similar functionality than self.jev, with the difference that it will take values from
            0 to 2*len(self) instead.
        '''
        # Check if given input is stored in the database. If not, re-evaluate.
        eval_required, kwargs_changed = self._eval_memcheck(*point, **kwargs)
        if eval_required or kwargs_changed:
            if warn:
                warnings.warn('Input parameter(s) changed; re-evaluating ...')
            self.eval(*point, compose=False, **kwargs)

        # In the non-periodic case the evaluation data might have to be extended.
        # Check if end-point agrees with the start.
        inp = self.dfunctions[self.ordering[0]]._input
        point = [inp[k].array(0)[0] for k in range(len(inp))] # the index [0] means we are looking at the first occurence 
        outp = self.dfunctions[self.ordering[-1]]._evaluation
        final_point = [outp[k].array(0)[-1] for k in range(len(outp))] # the index [-1] means we are looking at the last occurence
        periodic = periodic or all([check_zero(point[k] - final_point[k]) for k in range(len(point))]) # Use existing data in case that periodic == False, but start and end point agree nonetheless.

        # Define the function which will provide the jet-evaluation data
        L = len(self)
        if periodic:
            self._cycle = self
            jev = lambda k: self.jev(k%L)
        else:
            # In this scenario, data exists but the point is not fixed. Evaluation has to be done.
            if warn:
                warnings.warn(f"Evaluation data exists, but periodic: {periodic}. Extending & evaluating at\n {final_point}\n ...")
            self._cycle = self.__class__(*[copy(f) for f in self.dfunctions], order=self.order, ordering=self.ordering, run_params=self.run_params, reset=True)
            self._cycle.eval(*final_point, compose=False, **kwargs) # the final point will be the start point of the extension
            def jev(k):
                if k < L:
                    return self.jev(k)
                else:
                    return self._cycle.jev(k%L)
        self._cycle_periodic = periodic
        return jev
    
    def cycle(self, *args, outf='default', **kwargs):
        r'''
        Cycle through the given chain: Compute the derivatives at each point, assuming
        a periodic structure of the entire chain.
        
        The basic idea goes as follows: Let c1 --> c2 --> c3 --> c4 --> c5 denote the current
        chain (in this example consisting of 5 elements). Then this routine will extend the internal
        jet-evaluation data to numpy arrays of length 5 by suitable clones and identity operators,
        so that we can compute (compose) in parallel the following chain of length 2*5 - 1:
        
        |c1 ---> c2 ---> c3 ---> c4 ---> c5*
        |        c2 ---> c3 ---> c4 ---> c5 ---> c1*
        |                c3 ---> c4 ---> c5 ---> c1 ---> c2*
        |                        c4 ---> c5 ---> c1 ---> c2 ---> c3*
        |                                c5 ---> c1 ---> c2 ---> c3 ---> c4*
        
        The final results will then be returned as the starred values in the above diagram.
        
        Parameters
        ----------
        *point: float or iterable, optional
            The point at which to evaluate the derivative(s). It has to be provided if the chain has
            not yet stored any jet-evaluation data.
        
        periodic: boolean, optional
            If true, assume that the final point equals the given point. In particular, the
            data taken in the calculation will be taken from the current jet-evaluations (if
            it exists).
            
        outf: str, optional
            Output format; if 'default', compose the extended jet-evaluations and return a list of jet-evaluations,
            representing the derivatives along the chain.
            Otherwise, return a cderive object of length len(self)*2 - 1 for further processing. The
            object can be composed to yield the jet-evaluations along the current chain of len(self) in form of numpy
            arrays. This second option is intended to be used for performance improvements.
            
        **kwargs
            Optional keyworded arguments passed to a (possible) chain evaluation.
            
        Returns
        -------
        list or cderive
            Depending on the 'outf' parameter the output is either
            1) A list of jet-evaluations, where the k-th entry corresponds to the derivative of the
            chain from position k to position k + L, where L denotes the length of the chain (using
            a periodic chain structure).
            2) A cderive class with the property mentioned above.
        '''
        L = len(self)
        self._cycle_jev = self._prepare_cycle(*args, **kwargs)
        jev0 = self._cycle_jev(0)
        cycling_data = [[_jetp1(tile(jev0[component_index], ncopies=1), index=component_index, ncopies=L - 1) for component_index in range(len(jev0))]]
        
        # successively add copies and identity transformations to the given jet-evaluations, defining the traces to be tracked.
        for k in range(1, L):
            jevk = self._cycle_jev(k)
            cycling_data.append([_jetp1(tile(jevk[component_index], ncopies=k + 1), 
                                       index=component_index, 
                                       ncopies=L - k - 1) for component_index in range(len(jevk))])
            
        concat0 = [[jevk[ic].array(0) for ic in range(len(jevk))]]
        for k in range(L - 1):
            jevkpL = self._cycle_jev(k + L)
            cycling_data.append([_jetp1(tile(jevkpL[component_index], ncopies=L - k - 1), 
                                        to_concat0=[c[component_index] for c in concat0],
                                        index=component_index, 
                                        ncopies=k + 1, 
                                        location=0) for component_index in range(len(jevkpL))])
            concat0.append([jevkpL[ic].array(0) for ic in range(len(jevkpL))])
            
        self._cycle_data_inp = cycling_data
        if outf == 'default':
            # Compose the extended jet-evaluations and return the resulting list along the chain:
            self._cycle_data = compose(*cycling_data, run_params=self.run_params)
            return [[jcmp[k] for jcmp in self._cycle_data[k + L - 1]] for k in range(L)]
        else:
            # Construct a cderive class of length len(self)*2 - 1, having self._cycle_data_inp as jet-evaluations, and 
            # so that its compose routine will provide the same results as above.
            new_functions = [copy(f) for f in self.dfunctions] # shallow copy to prevent ._evaluation results to be overwritten in original in the code below
            ordering = (self.ordering*2)[:-1]
            for l in range(len(new_functions)):
                n_args = new_functions[l].n_args
                jets_l = [[self._cycle_data_inp[k][icmp] for k in range(len(ordering)) if ordering[k] == l] for icmp in range(n_args)]
                new_functions[l]._evaluation = [_jbuild(jc) for jc in jets_l]
            return self.__class__(*new_functions, ordering=ordering, order=self.order, run_params=self.run_params, reset=False)

    
def _get_ordering(sequence, start=0):
    '''
    For a given sequence of integers, find a unique ordering. E.g.
    ordering = [3, 4, 4, 7, 1]
    result = [0, 1, 1, 2, 3]
    
    Parameters
    ----------
    ordering: list
        The list of whose ordering we seek.
        
    start: int, optional
        An integer to control the start of the ordering.
        
    Returns
    -------
    list
        A list of integers describing the ordering of the given sequence.
    '''
    assert len(sequence) > 0
    new_ordering = []
    stash = []
    k = start
    for e in sequence:
        if e not in stash:
            new_ordering.append(k)
            k += 1
            stash.append(e)
        else:
            e_index = stash.index(e)
            new_ordering.append(e_index + start)
    return new_ordering

def _get_pattern_positions(sequence, pattern):
    '''
    For a given sequence and a given pattern, determine the positions at
    which the pattern occurs in the sequence -- non-overlapping -- starting
    from the first occurence in the sequence.
    
    Parameters
    ----------
    sequence: list
        A list of integers, representing the sequence.
        
    pattern: tuple
        A tuple of integers, representing the pattern.
    '''
    size = len(pattern)
    assert 1 <= size and size <= len(sequence)
    positions = []
    last_pos = -size
    k = 0
    for window in windowed(sequence, size):
        if window == pattern and last_pos + size <= k: # second condition ensures non-overlapping
            positions.append(k)
            last_pos = k
        k += 1
    return positions


# Tests of some of the examples in the documentation
import numpy as np

from njet import taylor_coefficients
from njet.extras import cderive
from njet.functions import sin, cos

def dequal(A, B, tol):
    # Convenience function to check if the values of two dictionaries agree up to a given tolerance
    if not A.keys() == B.keys():
        return False
    for k in A.keys():
        if not abs(A[k] - B[k]) < tol:
            return False
    return True

def test_docs_extras(tol=5e-15):

    g = 1
    per = lambda *z, **kwargs: [z[0], z[1] - g*z[1]**2]
    rot = lambda *z, alpha=0: [cos(alpha)*z[0] - sin(alpha)*z[1], sin(alpha)*z[0] + cos(alpha)*z[1]]

    z0 = [0.2, 0.1]
    
    M = 15
    ordering = [0, 1]*M
    drp = cderive(rot, per, order=2, ordering=ordering, n_args=2)
    
    alpha=1.22
    result1 = drp(*z0, alpha=alpha)
    
    ref1 = ({(0, 0): 0.18073767467258015,
             (1, 0): 0.3704449136384162,
             (0, 1): 0.3778714745973873,
             (0, 2): -1.2481054532394522,
             (2, 0): -2.5881230191830347,
             (1, 1): -0.74624063677712},
            {(0, 0): -0.012836565325653162,
             (1, 0): -0.4426107548482863,
             (0, 1): 0.8026126813486407,
             (0, 2): -0.49946504768391176,
             (2, 0): 0.7197445629662029,
             (1, 1): -0.3181227776491683})
    
    assert all([dequal(result1[k], ref1[k], tol=tol) for k in range(2)])
    
    result2 = drp.jev(4)[0].taylor_coefficients(n_args=2)
    
    assert result2 == {(0, 0): -0.0911100230191827,
                       (1, 0): 0.34364574631604705,
                       (0, 1): -0.9390993563190676}
    
    mysubchain = drp[1:13]
    result3 = mysubchain.jev(3)[0].taylor_coefficients(n_args=2)
    
    assert result3 == result2
    
    drpm = drp.merge(pattern=(1, 0, 1), positions=[1])
    assert len(drpm) == 28 and len(drp) == 30
    
    c1 = drp.compose()
    c2 = drpm.compose()

    result4 = taylor_coefficients(c1, n_args=2)
    assert result4 == result1
    
    result5 = taylor_coefficients(c2, n_args=2)
    assert all([dequal(result5[k], result1[k], tol=tol) for k in range(2)])
    
    cyc = drp.cycle(*z0, alpha=alpha)
    assert len(cyc) == len(drp)
    
    tc0 = taylor_coefficients(cyc[0], n_args=2)
    assert tc0 == ref1
        
    ordering2 = [1] + [0, 1]*(M - 1) + [0]
    drp2 = cderive(rot, per, order=2, ordering=ordering2, n_args=2)
    tc1 = taylor_coefficients(cyc[1], n_args=2)
    assert tc1 == drp2(*rot(*z0, alpha=alpha), alpha=alpha)
    
    z1 = np.array([[0.02, -0.056, z0[0]], [0.0031, 0.0118, z0[1]]])
    cyc_arr = drp.cycle(*z1, alpha=alpha)
    result6 = taylor_coefficients(cyc_arr[0], n_args=2)
    result7 = taylor_coefficients(cyc_arr[1], n_args=2)
    
    for k in range(2):
        for key in result6[k].keys():
            assert abs(result6[k][key][-1] - ref1[k][key]) < tol
            
    for k in range(2):
        for key in result7[k].keys():
            assert abs(result7[k][key][-1] - tc1[k][key]) < tol



def check_zero(value):
    # check if a value is zero; value may be an iterable
    check = value == 0
    if hasattr(check, '__iter__'):
        try:
            return check.all()
        except:
            return all(check)
    else:
        return check
    
def factorials(n: int):
    k = 1
    facts = [1]
    for j in range(1, n + 1):
        k *= j
        facts.append(k)
    return facts
    
def _nCr(n, r=[]):
    # Generator idea for n_over_k's,
    # taken from https://stackoverflow.com/questions/24093387/pascals-triangle-for-python
    for x in range(n + 1):
        l = len(r)
        r = [1 if i == 0 or i == l else r[i - 1] + r[i] for i in range(l + 1)]
        yield r
        
def nCr(n: int):
    return list(_nCr(n))

def convert_indices(list_of_tuples):
    '''
    Convert a list of tuples denoting the indices in a multivariate Taylor expansion 
    into a list of indices of the corresponding multilinear map.

    Parameters
    ----------
    list_of_tuples: list
        List of (self.n_args)-tuples denoting the indices in the multivariate Taylor expansion.

    Returns
    -------
    list
        List of tuples denoting the indices of a multilinear map.

    Example
    -------
    list_of_tuples = [(0, 2, 1, 9), (1, 0, 3, 0)]
    In this example we are looking at indices in the Taylor expansion of a function in 4 variables.
    The first member, (0, 2, 1, 9), corresponds to x0**0*x1**2*x2*x3**9 which belongs to the multilinear map
    of order 0 + 2 + 1 + 9 = 12, the second member to x0*x2**3, belonging to the multilinear map of order
    1 + 3 = 4.
    Hence, these indices will be transformed to tuples of length 12 and 4, respectively:
    (1, 1, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3) (variable 1 has power 2, variable 2 has power 1, variable 3 has power 9) 
    and
    (0, 2, 2, 2) (variable 0 has power 1, variable 2 has power 3).
    '''
    l1g = [[tpl[j]*[j] for j in range(len(tpl))] for tpl in list_of_tuples]
    return [tuple(e for subl in l1 for e in subl) for l1 in l1g] # flatten l1 and return list of converted tuples

def buildTruncatedJetFunction(*func, truncate=float('inf'), n_args: int=1):
    '''
    Modify a given chain of functions so that the output
    will be truncated between two function calls -- and at the end.
    
    Parameters
    ----------
    *func: callable(s)
        Functions which should be truncated. Note that these functions must support
        jets as input parameters.
    
    truncate: int, optional
        The power beyond which powers should be dropped.
        
    n_args: int, optional
        The number of input parameters of the series of functions.
        a) If n_args == 1, then it is assumed that those functions return
        individual jets/values. 
        b) If n_args > 1, it is assumed that *all* functions
        return iterables (vectors; their lengths may vary depending on the functions). 
        In case b) the user has to ensure that even for functions
        which take one argument, those functions return iterables of length 1.
        
    Returns
    -------
    callable
        A function taking n_args jet objects. 
    '''
    if n_args == 1: # we assume the output is not iterable here
        def tchain(z, **kwargs):
            for f in func:
                z = f(z, **kwargs)
                z = z.truncate(truncate)
            return z
    else:
        def tchain(*z, **kwargs):
            for f in func:
                z = f(*z, **kwargs)
                z = (*[ev.truncate(truncate) for ev in z],)
            return z
    return tchain


import numpy as np
import pytest

from njet import derive
from njet.functions import cos, sin
from njet.extras import general_faa_di_bruno, symtensor_call, cderive

def test_point_change(order=2):
    '''
    Basic test to confirm if derivatives are re-calculated if
    the input point was changed.
    '''
    f = lambda x, y: [x**2 + 0.42*x*y - y, 0.25*(x - y)**3]
    g = lambda x, y: [y**2 - 0.21*y - x**5, 0.883*y*x**2 + 0.32*y]
    
    z1 = [0.033, -1.13]
    z2 = [1.158, 0.332]
    
    ref1 = f(*g(*z1))
    ref2 = f(*g(*z2))
    
    dfg = cderive(g, f, order=order)
    r1 = dfg(*z1)
    r2 = dfg(*z2)
    
    assert all([r1[k][0, 0] == ref1[k] for k in range(2)])
    assert all([r2[k][0, 0] == ref2[k] for k in range(2)])

def test_symtensor_call():
    
    tol = 1e-15
    
    terms = {frozenset({(0, 1), (2, 1)}): 2.0,
             frozenset({(0, 2)}): 2.0,
             frozenset({(0, 0), (1, 2)}): -0.0003242141771667308}
    
    ref1 = 44.796693015392904
    ref2 = 5.9967578582283325
    
    result1 = symtensor_call([[7, 2, 4], [2.1, 5.1, 1.1]], terms)
    result2 = symtensor_call([[1, 2, 4], [2, 5, 1]], terms)

    assert abs(result1 - ref1) < tol
    assert abs(result2 - ref2) < tol

def test_general_faa_di_bruno():
    
    def f1(*x):
        return [x[0]**2 - (x[1] + 10 - 0.5*1j)**(-1) + x[2]*x[0], 1 + x[0]**2*x[1] + x[1]**3, x[2]]

    def g1(*x):
        return [(x[0] + x[1]*x[2]*(0.9*1j + 0.56) + 1)**(-3), 2*x[1]*4j + 5, x[0]*x[1]**6]
    
    start = [0.4, 1.67 + 0.01*1j, -3.5 + 2.1*1j]
    
    n = 7
    dg1 = derive(g1, order=n, n_args=3)
    df1 = derive(f1, order=n, n_args=3)
    
    evg1 = dg1.eval(*start)
    evf1 = df1.eval(*g1(*start))
    
    # Reference
    dfg1 = derive(lambda *z: f1(*g1(*z)), order=n, n_args=3)
    ref = dfg1.eval(*start)
    
    # General Faa di Bruno
    gen_fb = general_faa_di_bruno(evf1, evg1)
    
    tolerances = [[1e-15, 5e-15, 5e-15, 5e-14, 5e-14, 1e-12, 5e-12],
                  [1e-15, 5e-12, 1e-12, 5e-15, 5e-15, 1e-14, 2e-13],
                  [1e-15, 1e-15, 8e-14, 1e-15, 1e-15, 1e-15, 1e-15]]

    for k in range(3):
        diff = (gen_fb[k] - ref[k]).get_array()[1:]
        j = 0
        for e in diff:
            check = np.abs(list(e.terms.values()))
            if len(check) > 0:
                check = max(check)
            else:
                check = 0
            assert check < tolerances[k][j]
            j += 1
    
@pytest.mark.parametrize("x, y", [(0, 0), (0.56, 0.67), (np.array([0.1, -0.2]), np.array([0.526, 1.84]))])
def test_cderive1(x, y, phi=0.224*np.pi, n_reps: int=5, tol=1e-15):
    '''
    Test the derivatives of the 'n_reps'-times composition of a rotation,
    where the rotation takes an additional parameter (the rotation angle).
    '''
    assert n_reps >= 2
    ordering = [0]*n_reps
    rot = lambda *z, alpha=0: [cos(alpha)*z[0] - sin(alpha)*z[1], sin(alpha)*z[0] + cos(alpha)*z[1]]
    
    dcrot = cderive(rot, ordering=ordering, order=1, n_args=2)
    dref = derive(dcrot.jetfunc, order=1, n_args=2)
    ref = dref(x, y, alpha=phi/len(ordering))
    
    drot_phi = dcrot(x, y, alpha=phi/len(ordering))
    ref0 = rot(x, y, alpha=phi)
    refx = rot(1, 0, alpha=phi)
    refy = rot(0, 1, alpha=phi)
    
    # consistency checks
    if (0, 0) in drot_phi[0].keys():
        assert (np.abs(drot_phi[0][0, 0] - ref0[0]) < tol).all()
    if (0, 0) in drot_phi[1].keys():
        assert (np.abs(drot_phi[1][0, 0] - ref0[1]) < tol).all()
        
    assert (np.abs(drot_phi[0][1, 0] - refx[0]) < tol).all() # the derivative of the first component of rot in x-direction (== rotation x-component [0] in x-direction)
    assert (np.abs(drot_phi[0][0, 1] - refy[0]) < tol).all() # the derivative of the first component of rot in y-direction (== rotation x-component [0] in y-direction)

    assert (np.abs(drot_phi[1][1, 0] - refx[1]) < tol).all() # the derivative of the second component of rot in x-direction (== rotation y-component [1] in x-direction)
    assert (np.abs(drot_phi[1][0, 1] - refy[1]) < tol).all() # the derivative of the second component of rot in y-direction (== rotation y-component [1] in y-direction)
    
    dcrotm = dcrot.merge((0, 0), positions=[0, 3])
    result1 = dcrotm(x, y, alpha=phi/len(ordering))
    
    dcrotm_all = dcrot.merge()
    result2 = dcrotm_all(x, y, alpha=phi/len(ordering))
    
    dcrotm2 = dcrotm.merge()
    result3 = dcrotm2(x, y, alpha=phi/len(ordering))
    
    for result in [drot_phi, result1, result2, result3]:
        for k in range(2):
            assert all([(np.abs(ref[k][key] - result[k][key]) < tol).all() for key in ref[k].keys()])
    
ordering1 = [0, 1, 2, 3, 2, 1, 2, 4, 2, 1, 2, 4, 2, 1, 0]
pattern1_1 = (0, 1)
pattern1_2 = (2, 3, 2)
pattern1_3 = (1, 0)
pattern1_4 = (4, 2)

ordering2 = [0, 1, 2, 3, 2, 1, 2, 1, 2, 1, 4, 4, 2, 1, 2, 3, 2, 1, 0]
pattern2_1 = (2, 1, 2)
pattern2_2 = (4, 4)
pattern2_3 = (1, 2, 3)

point = [0.2, 0.14]

tolerances = {(0, 0): 1e-15, (1, 0): 5e-15, (0, 1): 1e-15, (0, 2): 5e-15, (2, 0): 1e-14, 
              (1, 1): 1e-14, (2, 1): 5e-13, (0, 3): 1e-13, (3, 0): 1e-12, (1, 2): 5e-13}

# Define a chain of vector-valued functions
f1 = [-0.41210319-0.93736294*1j, -0.12546181+0.99583435*1j, -0.05060304+0.72636541*1j, -0.57277969+0.7030248*1j]
g1 = [ 0.55362006+0.09802837*1j, -0.8305559 +0.49600788*1j,  0.03343595+0.58669087*1j, 0.94227846-0.58840718*1j]
h1 = [ 0.68077738+0.24660968*1j,  0.73098886+0.87092588*1j, -0.86900825-0.97546189*1j, -0.21817777+0.71363444*1j]

f2 = [-0.20674188-0.04716913*1j, -0.60197835-0.65897942*1j,  0.79173922-0.20115845*1j, -0.39323279-0.47489544*1j]
g2 = [ 0.25888809+0.74213104*1j, -0.46120294-0.45297721*1j, -0.68132361+0.36020504*1j, 0.36831401+0.8925468*1j ]
h2 = [ 0.67546745-0.25333062*1j, -0.24332418+0.01746218*1j, -0.1762223 +0.70213807*1j,-0.26571277-0.59505904*1j]

f3 = [ 0.76785636-0.27781605*1j,  0.46282915+0.27441113*1j, -0.49791468-0.57996221*1j, -0.48946897+0.62037892*1j]
g3 = [ 0.28804634-0.98764309*1j, -0.87406307+0.52595492*1j,  0.46174138+0.29365197*1j, -0.00551896-0.47692102*1j]
h3 = [-0.48132238-0.5247383*1j , -0.62638017+0.26876055*1j, -0.31827884+0.06467531*1j, 0.13965988-0.20589607*1j]

f4 = [-0.8096255 +0.96735152*1j,  0.06001575-0.31173884*1j, -0.69704779+0.33130702*1j, -0.1666179 +0.98778434*1j]
g4 = [-0.66590557+0.74032939*1j,  0.76306924+0.50447005*1j, -0.29480535+0.81453141*1j, -0.38647232+0.60297993*1j]
h4 = [-0.205739  +0.94352981*1j, -0.66742403+0.68390301*1j,  0.23363041-0.50400264*1j, 0.47711142-0.51922051*1j]

f5 = [-0.8096255 +0.96735152*1j,  0.06001575-0.31173884*1j, -0.69704779+0.33130702*1j, -0.1666179 +0.98778434*1j]
g5 = [-0.66590557+0.74032939*1j,  0.76306924+0.50447005*1j, -0.29480535+0.81453141*1j, -0.38647232+0.60297993*1j]
h5 = [-0.205739  +0.94352981*1j, -0.66742403+0.68390301*1j,  0.23363041-0.50400264*1j, 0.47711142-0.51922051*1j]

fl = [f1, f2, f3, f4, f5]
gl = [g1, g2, g3, g4, g5]
hl = [h1, h2, h3, h4, h5]

def opk(*z, f=[], g=[], h=[]):
    return [f[0]*z[0]**3 + f[1]*z[0]**2 + f[2]*z[0] + g[0]*z[1]**3 + \
            g[1]*z[1]**2 + g[2]*z[1], h[0]*z[0] + h[1]*z[1]]

opchain1 = [lambda *z, k=k: opk(*z, f=fl[k], g=gl[k], h=hl[k]) for k in range(5)] # "k=k" because of Python's late bindings, see https://stackoverflow.com/questions/49617380/list-comprehension-with-lambda-function

@pytest.mark.parametrize("pattern, ordering", [(pattern1_1, ordering1), (pattern1_2, ordering1), (pattern1_3, ordering1), 
                                               (pattern1_4, ordering1), (pattern2_1, ordering2), (pattern2_2, ordering2),
                                               (pattern2_3, ordering2)])
def test_cderive2(pattern, ordering, point=point, tolerances=tolerances, order=3):
    
    dopchain = cderive(*opchain1, ordering=ordering, order=order, n_args=2)
    
    cref = derive(dopchain.jetfunc, order=order, n_args=2)
    ref = cref(*point)#, disable_tqdm=disable_tqdm)
    
    r0 = dopchain(*point)

    dopm = dopchain.merge(pattern)
    r1 = dopm(*point)
    
    dopm2 = dopm.merge()
    r2 = dopm2(*point)
        
    for r in [r0, r1, r2]:
        for k in range(2):
            assert all([abs(ref[k][key] - r[k][key]) < tolerances[key] for key in ref[k].keys()])

def test_cderive3(point=[0.01, -0.053], order=3):
    '''
    Test if extracting two chains and putting their results together will give the same result
    as if we would derive (or compose) the entire chain directly.
    '''
    
    dopchain = cderive(*opchain1, ordering=ordering2, order=order, n_args=2)
    
    dr = derive(dopchain.jetfunc, order=order, n_args=2)
    
    _ = dopchain.eval(*point)
    
    part1 = dopchain[:10]
    part2 = dopchain[10:]
    
    part3 = part1.merge((2, 1))
    
    part4 = part2.merge((2, 1))
    part5 = part2.merge((4, 4))
    part6 = part2.merge((1, 2, 3))
    
    dr1 = part1.compose()
    dr2 = part2.compose()
    
    dr3 = part3.compose()
    
    dr4 = part4.compose()
    dr5 = part5.compose()
    dr6 = part6.compose()
    
    dr12 = general_faa_di_bruno(dr2, dr1)
    dr14 = general_faa_di_bruno(dr4, dr1)
    dr15 = general_faa_di_bruno(dr5, dr1)
    dr16 = general_faa_di_bruno(dr6, dr1)
    
    dr32 = general_faa_di_bruno(dr2, dr3)
    dr34 = general_faa_di_bruno(dr4, dr3)
    dr35 = general_faa_di_bruno(dr5, dr3)
    dr36 = general_faa_di_bruno(dr6, dr3)
    
    ref = dr.eval(*point)
    
    tolerances1 = [1e-15, 1e-15, 1e-14, 5e-14]
    tolerances2 = [1e-15, 1e-15, 1e-14, 2e-13]
    tolerances12 = [tolerances1, tolerances2]

    for r in [dr12, dr14, dr15, dr16, dr32, dr34, dr35, dr36]:
        for k in range(2):
            assert abs((ref[k] - r[k]).array(0)) < tolerances12[k][0] 
            for j in range(1, 4):
                assert  max(np.abs(list((ref[k] - r[k]).array(j).terms.values()))) < tolerances12[k][j]    

#################
# Cycling tests #
#################

def check_jet(A, tolerances: list):
    '''
    Convenience function to check if a jets containing jetpoly objects (as a result
    of a jet-evaluation) is close to zero.
    
    A list of tolerances will be used to compare the maximum of the 
    absolute values of the k-th array with its k-th entry.
    '''
    assert len(tolerances) == A.order + 1
    
    # Check the 0-parts:
    diff0 = np.abs(A.array(0))
    if hasattr(diff0, '__iter__'):
        if len(diff0) > 0:
            diff0 = np.max(diff0)
        else:
            diff0 = 0
    assert diff0 < tolerances[0], f'at 0: {diff0} >= {tolerances[0]}'
    
    # Check the higher-order parts:
    diff = A.get_array()[1:]
    j = 1
    for diffj in diff:
        check = np.abs(list(diffj.terms.values()))
        if len(check) > 0:
            check = np.max(check)
        else:
            check = 0
        assert check < tolerances[j], f'at {j}: {check} >= {tolerances[j]}'
        j += 1
                
cyc_point0 = (0, 0)
cyc_point1 = (0.0004, -0.0005)
                
cyc_ordering0 = [0, 1, 2, 3, 4]
cyc_ordering1 = [0, 1, 2, 3, 4, 3, 4, 3, 2, 3, 4, 3, 1, 0, 2, 0]
cyc_ordering2 = [1, 0, 2, 0, 1, 1, 4, 3, 2, 3, 4, 0, 1, 2, 2, 0, 4, 4, 3, 3]

cyc_tolerances0 = [1e-15, 1e-15, 1e-14, 5e-13]
cyc_tolerances1 = [1e-14, 2e-12, 5e-10, 1e-9]
cyc_tolerances2 = [1e-14, 2e-12, 5e-10, 1e-9]

@pytest.mark.parametrize("point, ordering, tolerances", [(cyc_point0, cyc_ordering0, cyc_tolerances0), 
                                                         (cyc_point0, cyc_ordering1, cyc_tolerances1),
                                                         (cyc_point0, cyc_ordering2, cyc_tolerances2),
                                                         (cyc_point1, cyc_ordering0, cyc_tolerances0),
                                                         (cyc_point1, cyc_ordering1, cyc_tolerances1),
                                                         (cyc_point1, cyc_ordering2, cyc_tolerances2)])
def test_cycling1(point, ordering, tolerances, order=3):
    '''
    Test if we receive the same results if we compute the derivatives of a chain
    separately or using the cycle feature.
    '''
    n_args = 2
    
    # Construct a series of chains, where every chain is cyclic shifted in comparison to the next:
    chains = []
    for k in range(len(ordering)):
        ordering_k = ordering[k:] + ordering[:k]
        chains.append(cderive(*opchain1, ordering=ordering_k, order=order, n_args=n_args))

    # To derive each chain separately we construct derive classes for the cyclic functions (cfunc's):
    def make_cfunc(k):
        c = chains[k]
        def cfunc(*z):
            for e in c:
                z = e.jetfunc(*z)
            return z
        return cfunc
    dchains = []
    for k in range(len(chains)):
        cfunc = make_cfunc(k) # the chain of functions
        dchains.append(derive(cfunc, order=order, n_args=n_args))

    # Compute the reference jet-evaluations (derivatives) at the point of interest:
    # First we have to make a function transporting the start point through the chain up to position k:
    def make_partf(k):
        functions = [opchain1[j] for j in ordering[:k]]
        def partf(*z):
            for f in functions:
                z = f(*z)
            return z
        return partf
    # Now compute the derivatives:
    refdirs = []
    for k in range(len(ordering)):
        partf_k = make_partf(k)
        refdirs.append(dchains[k].eval(*partf_k(*point)))
    
    # Construct the chain-derive class for the entire chain with respect to the given ordering and order:
    dopchain1 = cderive(*opchain1, ordering=ordering, order=order, n_args=n_args)
    # Consistency check: No evaluation results in this class should be taken over from anything previously:
    assert not hasattr(dopchain1, '_evaluation')
    assert not any([hasattr(dopchain1.dfunctions[k], '_evaluation') for k in range(len(opchain1))])
    
    # Cycle through the chain at the point of interest:
    cyc = dopchain1.cycle(*point)
    
    # Compare the results against the references:
    for k in range(len(ordering)):
        for cmp in range(n_args):
            check_jet(refdirs[k][cmp] - cyc[k][cmp], tolerances=tolerances[:order + 1])
        
        
cyc2_ordering = [0, 1, 3, 1, 2, 4, 0, 1, 0, 0, 0, 1, 1, 2, 2]
cyc2_tolerances = [1e-15, 5e-15, 5e-14, 5e-13, 5e-11]

cyc2_q1 = np.array([0.001, -0.0009])
cyc2_p1 = np.array([-0.0018, -0.0018])

cyc2_q2 = np.array([[0.00133, -0.000824], [0.00007, -0.00031]]) 
cyc2_p2 = np.array([[-0.00171, -0.00122], [0.000012, -0.000212]])

@pytest.mark.parametrize("q, p, ordering, tolerances", [(cyc2_q1, cyc2_p1, cyc2_ordering, cyc2_tolerances),
                                                        (cyc2_q2, cyc2_p2, cyc2_ordering, cyc2_tolerances)])
def test_cycling2(q, p, ordering, tolerances, order=4):
    '''
    Test if cycling works with numpy arrays.
    '''
    n_args = 2
    
    # build the chain-derive object
    dopchain1 = cderive(*opchain1, ordering=ordering, order=order, n_args=n_args)
    
    # derive the chain of functions directly to provide a reference
    df = derive(dopchain1.jetfunc, order=order, n_args=n_args)
    ref = df.eval(q, p)
    
    # verify that dopchain1 has not yet any evaluation results stored
    assert not hasattr(dopchain1, '_evaluation')
    assert not any([hasattr(f, '_evaluation') for f in dopchain1.dfunctions])
    
    cyc = dopchain1.cycle(q, p)
    
    for j in range(n_args):
        check_jet(ref[j] - cyc[0][j], tolerances=tolerances)
        
        
cyc3_point1 = 0.0003, 0.00063
cyc3_point2 = -0.00037, 0.00049
cyc3_point3 = np.array([[cyc3_point1[0], cyc3_point2[0]], [cyc3_point1[1], cyc3_point2[1]]])

cyc3_ordering = [0, 1, 0, 1, 0, 2, 0, 1, 0, 2, 2, 0, 0, 2, 0, 0, 1, 4, 3, 2, 4]
cyc3_tolerances = [1e-15, 5e-15, 1e-13, 1e-11]

@pytest.mark.parametrize("point, ordering, tolerances", [(cyc3_point1, cyc3_ordering, cyc3_tolerances),
                                                         (cyc3_point2, cyc3_ordering, cyc3_tolerances),
                                                         (cyc3_point3, cyc3_ordering, cyc3_tolerances)])
def test_cycling3(point, ordering, tolerances, order=3):
    '''
    Test cycling if returning a cderive object:
    1) Comparison to default cycling.
    2) Merging & Comparison.
    3) Using numpy arrays. (Note that in a previous test, the result of using numpy arrays with the 'default' outf-parameter
       already agrees with the single-point results, so the test on equality between the 'default' array-result to the
       array-result for cderive objects should imply also agreement with the single-point results.)
    '''
    
    n_args = 2
    d_ref = cderive(*opchain1, order=order, ordering=ordering, n_args=n_args)
    d1 = cderive(*opchain1, order=order, ordering=ordering, n_args=n_args)

    cyc_ref = d_ref.cycle(*point, outf='default')
    dcyc = d1.cycle(*point, outf=None)

    dcycm = dcyc.merge(pattern=(ordering[2], ordering[3]))

    cyc = dcyc.compose()
    cycm = dcycm.compose()
        
    for j in range(2):
        tc_j = cyc[j]
        tcm_j = cycm[j]
            
        for k in range(len(ordering)):
            tc_ref_j = cyc_ref[k][j]
            check_jet(tc_j[k] - tc_ref_j, tolerances=tolerances)
            check_jet(tcm_j[k] - tc_ref_j, tolerances=tolerances)
            
        


import pytest

from njet import derive, jet
from njet.functions import sin, cos, exp
from njet.common import buildTruncatedJetFunction

import sympy
from sympy import Symbol
import numpy as np

######################
# Basic derivatives
######################

@pytest.mark.parametrize("point", (0, 0.32, 1.55, -1.268))
def test_cos_sin(point, order=9, tol=1e-15):
    dcos = derive(cos, order)
    dsin = derive(sin, order)

    tc_cos = dcos(point)
    tc_sin = dsin(point)
    
    c = np.cos(point)
    s = np.sin(point)
    
    for k in range(order):
        if k%4 == 0:
            c_expectation = c
            s_expectation = s
        elif k%4 == 1:
            c_expectation = -s
            s_expectation = c
        elif k%4 == 2:
            c_expectation = -c
            s_expectation = -s
        else:
            c_expectation = s
            s_expectation = -c
        
        if c_expectation != 0:
            assert abs(tc_cos[(k,)] - c_expectation) < tol
        if s_expectation != 0:
            assert abs(tc_sin[(k,)] - s_expectation) < tol

######################
# Chained expressions
######################

def test_ce1():
    H = lambda x, y, px, py: 0.24*(x**2 + px**2) + 0.86*(y**2 + py**2) + x**3 - y**3 + 9*x**4*y + sin(x)
    dH = derive(H, order=5)
    
    z = [1.1, 0.14, 1.26, 0.42]
    z0 = [0, 0, 0, 0]
    
    H_shift = lambda x, y, px, py: H(x + z[0], y + z[1], px + z[2], py + z[3])
    dH_shift = derive(H_shift, order=5)
    
    hdict = dH(*z)
    hshiftdict = dH_shift(*z0)
    
    check1 = all([hdict[k] == hshiftdict[k] for k in hdict.keys()])
    
    hdict2 = dH(*z)
    hshiftdict2 = dH_shift(*z0)

    check2 = all([hdict2[k] == hshiftdict2[k] for k in hdict2.keys()])
    test_failed = not (check1 and check2)
    if test_failed:
        print ('Shift test failed.')
    else:
        print ('Shift test succeeded.')
    assert not test_failed
    
    
@pytest.mark.parametrize("x0", (1.55, 0.42, -1.631))
def test_composition_1d(x0, tol=5e-12):
    '''
    Test if the composition @ of two jets agrees with the expectation (in the 1D case).
    '''
    f = lambda x: x**3 + 2*x
    f1 = lambda x: 3*x**2 + 2
    f2 = lambda x: 6*x
    f3 = lambda x: 6

    g = lambda x: 1/x
    g1 = lambda x: -1*x**-2
    g2 = lambda x: 2*x**-3
    g3 = lambda x: -6*x**-4

    # the expectation
    fg = lambda x: x**-3 + 2/x
    fg1 = lambda x: -3*x**-4 - 2*x**-2
    fg2 = lambda x: 12*x**-5 + 4*x**-3
    fg3 = lambda x: -60*x**-6 - 12*x**-4
     
    y0 = g(x0)
    fjet = jet(f(y0), f1(y0), f2(y0), f3(y0))
    gjet = jet(g(x0), g1(x0), g2(x0), g3(x0))
    fgjet = jet(fg(x0), fg1(x0), fg2(x0), fg3(x0))
    diff = fgjet - fjet@gjet
    
    assert all([abs(e) < tol for e in diff.get_array()])
    
    
######################## 
# Symbolic differentiation
########################
def test_sd1():
    d2 = derive(lambda x, y: sin(-x**2 + y), order=4)
    x, y = Symbol('x'), Symbol('y')
    K = d2(x, y)
    a = x**2 - y
    c, s = sympy.cos(a), sympy.sin(a)
    Kref = {}
    Kref[(0, 0)] = -s
    Kref[(1, 0)] = -2*x*c
    Kref[(0, 1)] = c
    Kref[(2, 0)] = 4*x**2*s - 2*c
    Kref[(1, 1)] = -2*x*s
    Kref[(0, 2)] = s
    Kref[(2, 1)] = -4*x**2*c - 2*s
    Kref[(0, 3)] = -c
    Kref[(3, 0)] = 8*x**3*c + 12*x*s
    Kref[(1, 2)] = 2*x*c
    Kref[(0, 4)] = -s
    Kref[(2, 2)] = -4*x**2*s + 2*c
    Kref[(3, 1)] = 8*x**3*s - 12*x*c
    Kref[(1, 3)] = 2*x*s
    Kref[(4, 0)] = -16*x**4*s + 48*x**2*c + 12*s
    test_failed = False
    for k in K.keys():
        check_k = sympy.nsimplify(K[k].expand() - Kref[k].expand()) # use of nsimplify due to: https://stackoverflow.com/questions/64761602/how-to-get-sympy-to-replace-ints-like-1-0-with-1
        if check_k != 0:
            print (k, check_k)
            test_failed = True
    if test_failed:
        print ('Symbolic differentiation test 1 failed.')
    else:
        print ('Symbolic differentiation test 1 succeeded.')
    assert not test_failed


def test_sd2():
    d2 = derive(lambda x, y: -sin(x**2 - y), order=4)
    df = lambda x, y: d2(x, y)[(2, 1)] # dx dx dy - component of the function above.
    df_nested = derive(df, order=3)
    x, y = Symbol('x'), Symbol('y')
    a = x**2 - y
    c, s = sympy.cos(a), sympy.sin(a)
    K2 = df_nested(x, y)
    Kref2 = {}
    Kref2[(0, 0)] = -4*x**2*c - 2*s
    Kref2[(1, 0)] = 8*x**3*s - 12*x*c
    Kref2[(0, 1)] = -4*x**2*s + 2*c
    Kref2[(2, 0)] = 16*x**4*c + 8*x**2*s + 40*x**2*s - 12*c
    Kref2[(1, 1)] = -8*x**3*c - 12*x*s
    Kref2[(0, 2)] = 4*x**2*c + 2*s
    Kref2[(2, 1)] = 16*x**4*s - 8*x**2*c - 40*x**2*c - 12*s
    Kref2[(0, 3)] = 4*x**2*s - 2*c
    Kref2[(3, 0)] = -32*x**5*s + 160*x**3*c + 120*x*s
    Kref2[(1, 2)] = -8*x**3*s + 12*x*c
    test_failed = False
    for k in K2.keys():
        check_k = sympy.nsimplify(K2[k].expand() - Kref2[k].expand())
        if check_k != 0:
            print (k, check_k)
            test_failed = True
    if test_failed:
        print ('Symbolic differentiation test 2 failed.')
    else:
        print ('Symbolic differentiation test 2 succeeded.')
    assert not test_failed


def test_sd3():
    xmpl2 = lambda a, b, c: sin(a**5 + b**2*c)
    dd = derive(xmpl2, order=3)
    a, b, c = Symbol('a'), Symbol('b'), Symbol('c')
    K3 = dd(a, b, c)
    arg = a**5 + b**2*c
    si = sympy.sin(arg)
    co = sympy.cos(arg)
    Kref3 = {}
    Kref3[(0, 0, 0)] = si
    Kref3[(0, 0, 1)] = b**2*co
    Kref3[(1, 0, 0)] = 5*a**4*co
    Kref3[(0, 1, 0)] = 2*b*c*co
    Kref3[(0, 2, 0)] = -4*b**2*c**2*si + 2*c*co
    Kref3[(0, 0, 2)] = -b**4*si
    Kref3[(1, 0, 1)] = -5*a**4*b**2*si
    Kref3[(0, 1, 1)] = -2*b**3*c*si + 2*b*co
    Kref3[(2, 0, 0)] = -25*a**8*si + 20*a**3*co
    Kref3[(1, 1, 0)] = -10*a**4*b*c*si
    Kref3[(2, 1, 0)] = -50*a**8*b*c*co - 40*a**3*b*c*si
    Kref3[(1, 1, 1)] = -10*a**4*b**3*c*co - 10*a**4*b*si
    Kref3[(0, 1, 2)] = -2*b**5*c*co - 4*b**3*si
    Kref3[(0, 2, 1)] = -4*b**4*c**2*co - 10*b**2*c*si + 2*co
    Kref3[(0, 3, 0)] = -8*b**3*c**3*co - 12*b*c**2*si
    Kref3[(3, 0, 0)] = -125*a**12*co - 300*a**7*si + 60*a**2*co
    Kref3[(2, 0, 1)] = -25*a**8*b**2*co - 20*a**3*b**2*si
    Kref3[(0, 0, 3)] = -b**6*co
    Kref3[(1, 2, 0)] = -20*a**4*b**2*c**2*co - 10*a**4*c*si
    Kref3[(1, 0, 2)] = -5*a**4*b**4*co
    test_failed = False
    for k in K3.keys():
        check_k = sympy.nsimplify(K3[k].expand() - Kref3[k].expand())
        if check_k != 0:
            print (k, check_k)
            test_failed = True
    if test_failed:
        print ('Symbolic differentiation test 3 failed.')
    else:
        print ('Symbolic differentiation test 3 succeeded.')
    assert not test_failed


########################
# Nested differentiation
########################
def test_nd1(tol=1e-12, verbose=False):
    f = lambda x, y: sin(1/x + y)
    df = derive(f, order=5)
    print ('Nested derivative test 1 ...')
    zref=[0.2, 1.1]
    nested_test_failed = False
    D = df(*zref)
    for a, b in [(0, 1), (1, 0), (0, 2), (1, 1), (2, 0), (0, 3), (1, 2), (2, 1), (3, 0)]:
        dabf = lambda x, y: df(x, y)[(a, b)]
        ddabf = derive(dabf, order=2)
        D2 = ddabf(*zref)
        if verbose:
            print (f'--- D{a}{b} ---')
        for key in D2.keys():
            key2 = (key[0] + a, key[1] + b)
            rel_error = abs(D[key2] - D2[key])/D[key2]
            check = rel_error < tol
            if verbose:
                print ( rel_error, check )
            if check == False:
                nested_test_failed = True
        if verbose:
            print ('-----------')

    if nested_test_failed:
        print (f'Nested derivative test 1 failed with given tolerance {tol}.')
    else:
        print (f'Nested derivative test 1 succeeded with given tolerance {tol}.')
    assert not nested_test_failed

def test_nd2(tol=1e-12, verbose=False):
    f = lambda x, y: sin(1/x + y)
    df = derive(f, order=3)
    dxxf = lambda x, y: df(x, y)[(1, 1)] 
    dxxyf = lambda x, y: df(x, y)[(2, 1)]
    g = lambda x, y: f(x, y)/(1 + dxxf(x, y)) + dxxyf(x, y)**-3
    dg = derive(g, order=12)

    print ('Nested derivative test 2 ...')
    lin = np.linspace(0.02, 0.32, 11)
    nested_test2_failed = False
    try:
        res = dg(0.2, 1.1)
        res2 = dg(lin, 1.1)
    except:
        nested_test2_failed = True
        
    assert not nested_test2_failed

    for key in res.keys():
        rel_error = abs(res2[key][6] - res[key])/res[key]
        if rel_error > tol:
            if verbose:
                print (f'problem at key: {key}, {rel_error}')
            nested_test2_failed = True

    if nested_test2_failed:
        print (f'Nested derivative test 2 failed with given tolerance {tol}.')
    else:
        print (f'Nested derivative test 2 succeeded with given tolerance {tol}.')
    assert not nested_test2_failed
    
    
def test_nd3():
    def prime(f, k=0):
        df = derive(f, order=1)
        return lambda x: df.grad(x)[(k,)]
    xmpl4 = lambda x: sin(x**2)
    x = Symbol('x')
    c, s = sympy.cos(x**2), sympy.sin(x**2)
    d3f = prime(prime(prime(xmpl4)))(x)
    check = sympy.nsimplify(d3f.expand() - (-8*x**3*c - 12*x*s))
    nested_test3_failed = True
    if check == 0:
        nested_test3_failed = False
    if nested_test3_failed:
        print ('Nested derivative test 3 failed.')
    else:
        print ('Nested derivative test 3 succeeded.')
        
    assert not nested_test3_failed
    
######################## 
# Projection
########################
def test_projection1():
    f1 = lambda *z: z[0] + z[1]**3
    df1 = derive(f1, n_args=2, order=3)
    x = Symbol('x')
    y = Symbol('y')
    out1 = df1(x, 2*y)
    out1_ref = {(0, 0): x + 8*y**3, (1, 0): 1, (0, 1): 12*y**2, (0, 2): 12*y, (0, 3): 6}
    assert out1.keys() == out1_ref.keys()
    for k, v in out1.items():
        assert out1_ref[k] - v == 0

def test_projection2():
    f1 = lambda *z: z[0] + z[1]**3
    df1 = derive(f1, n_args=2, order=3)
    out2 = df1(np.array([2, 1]), np.array([2, 8]))
    out2_ref = {(0, 0): np.array([10., 513.]), (1, 0): np.array([1.0, 1.0]), (0, 1): np.array([12., 192.]), 
                (0, 2): np.array([12., 48.]), (0, 3): np.array([6., 6.])}
    for k, v in out2.items():
        assert (out2_ref[k] == v).all()
        assert v.shape == out2_ref[k].shape
    
def test_projection3():
    f3 = lambda z: z[0]*z[1]
    df3 = derive(f3, n_args=1, order=3)
    out3 = df3(np.array([2, 1, 5]))
    ref3 = {(0,): 2, (1,): 3, (2,): 2}
    for k in ref3.keys():
        assert ref3[k] == out3[k]

def test_projection4():
    f4 = lambda z, w, y: z[0] + z[1]**3 + z[2]*z[0] + y[0]*w[2]
    df4 = derive(f4, n_args=3, order=3)
    
    z0, w0, y0 = np.array([2, 1, 5]), np.array([4, 2, 1]), np.array([-1, -2, 2])
    out3_a = df4(z0, w0, y0)
    out3_b = df4(z0, w0, y0, mult_prm=False, mult_drv=False)
    
    ref_a = {(0, 0, 0): 12, 
             (0, 0, 1): 1,
             (1, 0, 0): 11,
             (0, 1, 0): -1,
             (0, 1, 1): 1,
             (2, 0, 0): 8,
             (3, 0, 0): 6}
    
    ref_b = {(0, 0, 0): 12, 
             (0, 0, 1): 1,
             (1, 0, 0): 11,
             (0, 1, 0): -1,
             (0, 1, 1): 2,
             (2, 0, 0): 8,
             (3, 0, 0): 6}
    
    for k in ref_a.keys():
        assert ref_a[k] == out3_a[k]
        assert ref_b[k] == out3_b[k]
        
def test_projection5():
    f5 = lambda z: z[0, 0] + z[0, 1]**3 + z[0, 2]*z[0, 0] + z[1, 0]*z[2, 2]
    df5 = derive(f5, n_args=1, order=3)
    out5 = df5(np.array([[2, 1, 5], [4, 2, 1], [-1, -2, 2]]))
    ref5 = {(0,): 21, (1,): 17, (2,): 10, (3,): 6}
    for k in ref5.keys():
        assert ref5[k] == out5[k]
        
def test_projection6():
    f6 = lambda z: z[0, :] + z[:, 1]**3 + z[0, :]*z[:, 0] + z[1, 0]*z[2, 2]
    df6 = derive(f6, n_args=1, order=3)
    out6 = df6(np.array([[2, 1, 5], [4, 2, 1], [-1, -2, 2]]))
    ref6 = {(0,): np.array([15., 21.,  0.]), (1,): np.array([14., 24., 23.]), 
            (2,): np.array([10., 16., -8.]), (3,): np.array([6., 6., 6.])}
    for k in ref6.keys():
        assert len(ref6[k]) == len(out6[k])
        assert (ref6[k] == out6[k]).all()

########################
# Differentiation of vector-valued functions
########################

def test_2d_function():
    f2d = lambda x, y: [x**2 + y, x*y + y**3]
    df2d = derive(f2d, n_args=2, order=5)
    x0, y0 = 4, 9
    out2d = df2d(x0, y0)
    ref2d = ({(0, 0): 25.0, (1, 0): 8.0, (0, 1): 1.0, (2, 0): 2.0},
             {(0, 0): 765.0, (1, 0): 9.0, (0, 1): 247.0, (1, 1): 1.0, (0, 2): 54.0, (0, 3): 6.0})
    for j in range(2):
        for k in out2d[j].keys():
            assert out2d[j][k] == ref2d[j][k]


########################
# Truncation
########################
@pytest.mark.parametrize("point, truncate", [(0.23, 1), (0.23, 3), (np.array([0.35, 1.55]), 1), (np.array([0.35, 1.55]), 4)]) 
def test_truncate_1d(point, truncate, order=7):
    '''
    Test if truncation gives the low-order derivatives
    '''
    chain = [cos, sin, lambda x: sin(x)**2, lambda x: x**2 - 3]
    
    def reference_func(z):
        for c in chain:
            z = c(z)
        return z
    dref = derive(reference_func, order=order, n_args=1)
    
    truncated_chain = buildTruncatedJetFunction(*chain, truncate=truncate, n_args=1)
    dtrunc = derive(truncated_chain, order=order, n_args=1)
    
    reference = dref(point)
    result = dtrunc(point)
    for k in result.keys():
        assert (np.array(result[k] - reference[k]) == 0).all()
        


# -*- coding: utf-8 -*-

# DO NOT EDIT THIS FILE!
# This file has been autogenerated by dephell <3
# https://github.com/dephell/dephell

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


def get_long_description():
    with open(f'./README.md', encoding='utf8') as fp:
        return fp.read()
        
setup(
    long_description=get_long_description(),
    name='njet',
    version='0.4.9',
    description='Lightweight automatic differentiation package for higher-order differentiation.',
    long_description_content_type='text/markdown',
    python_requires='<3.11,>=3.8',
    project_urls={
        "homepage": "https://njet.readthedocs.io/en/latest/index.html",
        "repository": "https://github.com/mtitze/njet"
    },
    author='Malte Titze',
    author_email='mtitze@users.noreply.github.com',
    license='GPL-3.0-or-later',
    keywords='AD jet njet automatic differentiation',
    packages=['njet'],
    package_dir={"": "."},
    package_data={},
    install_requires=[
        'numpy==1.*,>=1.21.2', 'sympy==1.*,>=1.8.0',
        'more_itertools==9.*,>=9.0.0', 'tqdm==4.*,>=4.64.0'
    ],
    extras_require={"dev": ["ipykernel==6.*,>=6.4.1"]},
)



from .functions import exp, log, zero
from .jet import jet as _jet
from .jet import jetpoly
from .common import check_zero, convert_indices

class jet(_jet):
    '''
    Class to store and operate with higher-order derivatives of a given function.
    '''
    def __pow__(self, other):
        other = self.convert(other)

        if other.order == 0:
            result = _jet.__pow__(self, other)
        else:
            '''
            General exponentiation, using exp and log functions.
            '''
            result = exp(log(self)*other)
            result.graph = [(2, '**'), self.graph, other.graph]
        return result
    
def getNargs(func):
    '''
    Determine the number of arguments of a function. Raise an error if they can not be determined.
    
    Parameters
    ----------
    func: callable
        The function to be examined.
        
    Returns
    -------
    int
        The number of arguments of the given function.
    '''
    error_msg = 'The number of function arguments could not be determined.'
    n_args = 0
    try:
        n_args = func.__code__.co_argcount
    except:
        raise RuntimeError(error_msg)
    assert n_args > 0, error_msg
    return n_args

def identity(*z, order: int):
    '''
    Create a set of jets representing the components of the identity operator.
    
    Parameters
    ----------
    *z
        The point at which the identitity is evaluated.
        
    order: int, optional
        The order of the desired jets.
        
    Returns
    -------
    list
        A list of jets, representing the components of the identity.
    '''
    return [jet(z[k], jetpoly(zero(z[k]) + 1, index=k, power=1), n=order) for k in range(len(z))] # add a zero to jetpoly values to produce the same shape or type as z[k]

def taylor_coefficients(evaluation, output_format=0, **kwargs):
    '''
    Return the Taylor coefficients of a jet evaluation.
    
    Parameters
    ----------
    output_format: int, optional
        If 0 and the output would be a list of length 1, return the entry 
        of this list instead. If != 0, disable this behavior.
    
    **kwargs
        Parameters passed to njet.jet.taylor_coefficients routine.
        
    Returns
    -------
    dict
        One or more dictionaries, representing the Taylor-coefficients of the given
        evaluation (see njet.poly.jetpoly.taylor_coefficients for details).
    '''
    try:
        out = evaluation.taylor_coefficients(**kwargs)
        if output_format != 0:
            out = [out]
    except:
        out = (*[ev.taylor_coefficients(**kwargs) for ev in evaluation],) # also stored in ev._tc
        if len(out) == 1 and output_format == 0:
            out = out[0]
    return out

class derive:
    '''
    Class to handle the derivatives of a (jet-)function (i.e. a function consisting of a composition
    of elementary functions).
    
    Parameters
    ----------
    func: callable(s)
        The function(s) to be derived. Must be expressed in terms of functions (e.g. polynomials) supported by njet.functions. Note that the first function in the given chain will be executed first.

    order: int, optional
        The order up to which the function should be derived.

    n_args: int, optional
        The number of arguments on which func depends on; passed to getNargs routine.
        
    n_out: int, optional
        Define the number of output parameters of the function(s). This is required for multi-dimensional
        output.
        
    truncate: int, optional
        If given, truncate the jets after each iteration through the given functions.
    '''
    def __init__(self, func, order: int=1, n_args: int=0):
        self.order = order
        self.jetfunc = func
        
        # Determine the number of input parameters
        if n_args == 0:
            self.n_args = getNargs(func)
        else:
            self.n_args = n_args

    def jet_input(self, *z):
        inp = identity(*z, order=self.order)
        self._input = inp
        return inp
        
    def eval(self, *z, **kwargs):
        '''
        Pass a jet of order self.order, having polynomials in its higher-order entries,
        through the given function.
        
        Parameters
        ----------
        z: subscriptable
            List of values at which the function and its derivatives should be evaluated.
            
        **kwargs:
            Optional keyword arguments passed to the underlying function.
                    
        Returns
        -------
        jet
            Jet containing the value of the function in its zero-th entry and the
            jet-polynomials in the higher-order entries.
        '''
        self._evaluation = self.jetfunc(*self.jet_input(*z), **kwargs)
        return self._evaluation
            
    def __call__(self, *z, **kwargs):
        '''Evaluate the derivatives of a (jet-)function at a specific point up to self.order.
        
        Parameters
        ----------
        z: subscriptable
            List of values at which the function and its derivatives should be evaluated.
            
        **kwargs:
            Optional arguments passed to self._evaluation.taylor_coefficients routine.

        Returns
        -------
        dict
            Dictionary of compontens of the multivariate Taylor expansion of the given function self.jetfunc
        '''
        # These two keywords are reserved for the taylor_coefficients routine and will be removed from the input:
        mult_prm = kwargs.pop('mult_prm', True)
        mult_drv = kwargs.pop('mult_drv', True)
        # perform the computation, based on the input vector
        return taylor_coefficients(self.eval(*z, **kwargs), n_args=self.n_args, mult_prm=mult_prm, mult_drv=mult_drv)
        
    def build_tensor(self, k: int, **kwargs):
        '''
        Convert the components of the k-th derivative into the entries of a (self.n_args)**k tensor.
        See njet.jet.build_tensor for details.
        '''
        assert k <= self.order, f'Order ({self.order}) insufficient for requested number ({k}) of derivatives.'
        assert hasattr(self, '_evaluation'), 'Derivative(s) need to be evaluated first. Try passing a point.'
        return self._evaluation.build_tensor(k=k, **kwargs)
    
    def grad(self, *z, **kwargs):
        '''
        Returns the gradient of the current function.
        See njet.jet.grad for details.
        '''
        if len(z) > 0:
            _ = self(*z, **kwargs)
        return self.build_tensor(k=1, **kwargs)
    
    def hess(self, *z, **kwargs):
        '''
        Returns the Hessian of the function.
        See njet.jet.hess for details.
        '''
        if len(z) > 0:
            _ = self(*z, **kwargs)
        return self.build_tensor(k=2, **kwargs)
        


import os
import numpy as np

from njet.functions import log
from njet import derive

# Examples related from issue 11; this should go through without errors


######### Test1

xfilename = os.path.join(os.path.dirname(__file__), 'example1.npy')
x = np.load(xfilename)

def grad(fun):
    "A decorator to evaluate the gradient."
    
    def inner(x, **kwargs):
        # init the output
        out = np.zeros((9, *x.shape[2:]))
        
        # obtain the gradient and loop over items
        for key, value in fun.grad(*flatten(x), **kwargs).items():
            out[key] = value
        
        # reshape the gradient
        return out.reshape(3, 3, -1)
    
    return inner

def hess(fun):
    "A decorator to evaluate the hessian."

    def inner(x, **kwargs):
        # init the output
        out = np.zeros((9, 9, *x.shape[2:]))
        
        # obtain the hessian and loop over items
        for key, value in fun.hess(*flatten(x), **kwargs).items():
            out[key] = value
        
        # reshape the hessian
        return out.reshape(3, 3, 3, 3, -1)
    
    return inner

def trace(x):
    return x[0][0] + x[1][1] + x[2][2]

def det(x):
    a = x[0][0] * x[1][1] * x[2][2]
    b = x[0][1] * x[1][2] * x[2][0]
    c = x[0][2] * x[1][0] * x[2][1]
    d = x[2][0] * x[1][1] * x[0][2]
    e = x[2][1] * x[1][2] * x[0][0]
    f = x[2][2] * x[1][0] * x[0][1]
    return a + b + c - d - e - f

flatten = lambda x: x.reshape(9, -1)
reshape = lambda x: [[x[0], x[1], x[2]], [x[3], x[4], x[5]], [x[6], x[7], x[8]]]
# reshape = lambda x: np.array(x, dtype=object).reshape(3, 3, -1) # this fails

def tensorjet(fun):
    "A decorator to reshape the input."
    def inner(*x, **kwargs):
        return fun(reshape(x), **kwargs)
    return inner

@tensorjet
def fun(C):
    "The Neo-Hookean material formulation (isotropic hyperelasticity)."
    return trace(C) - log(det(C))

def test_examples1():
    '''
    Test issue 11 example
    '''
    dfun = derive(fun, order=2, n_args=9)
    dfundx = grad(dfun)(x)
    d2fundx2 = hess(dfun)(x)
    
    r1filename = os.path.join(os.path.dirname(__file__), 'example1_result1.npy')
    r1 = np.load(r1filename)
    r2filename = os.path.join(os.path.dirname(__file__), 'example1_result2.npy')
    r2 = np.load(r2filename)

    assert dfundx.shape == (3, 3, 2**4)
    assert d2fundx2.shape == (3, 3, 3, 3, 2**4)
    
    assert (dfundx == r1).all()
    assert (d2fundx2 == r2).all()

######### Test2

def trace2(x):
    return x[0, 0] + x[1, 1] + x[2, 2]

def det2(x):
    a = x[0, 0] * x[1, 1] * x[2, 2]
    b = x[0, 1] * x[1, 2] * x[2, 0]
    c = x[0, 2] * x[1, 0] * x[2, 1]
    d = x[2, 0] * x[1, 1] * x[0, 2]
    e = x[2, 1] * x[1, 2] * x[0, 0]
    f = x[2, 2] * x[1, 0] * x[0, 1]
    return a + b + c - d - e - f

def fun2(C):
    return trace2(C) - log(det2(C))

def test_examples2():
    '''
    Test a similar function as in the issue 11 example, but now where the 
    argument is assumed to contain a tensor variable. The result
    will not make sense in the original context; the test is only here to
    ensure that a result is produced.
    '''
    dfun2 = derive(fun2, order=2, n_args=1)
    y = dfun2(x)
    
    for k in range(3):
        rkfilename = os.path.join(os.path.dirname(__file__), f'example2_result_{k}.npy')
        ref_k = np.load(rkfilename)
        assert (ref_k == y[(k,)]).all()
        
    
if __name__ == '__main__':
    test_examples1()
    test_examples2()

